
from datetime import datetime, timezone

def utc_now():
    return datetime.utcnow().replace(tzinfo=timezone.utc)

class TraceInProgress:
    session_id: str
    total_events: int
    started_at: datetime

    def __init__(self, session_id: str):
        self.session_id = session_id
        self.total_events = 0
        self.started_at = utc_now()

    def add_events(self, events_count):
        self.total_events += events_count



from datetime import datetime
from pathlib import Path
from typing import List, Set, Any

from pycrunch_trace.events.base_event import Event
from pycrunch_trace.file_system.session_store import SessionStore




class TraceSession:
    recording_directory : Path
    files_in_session: Set[str]
    excluded_files : Set[str]
    environment_during_run: dict
    event_buffer: List[Any]

    def __init__(self):
        self.files_in_session = set()
        self.excluded_files = set()
        self.environment_during_run = dict()
        self.session_store = SessionStore()

    def buffer_became_available(self, event_buffer):
        self.event_buffer = event_buffer

    def did_enter_traceable_file(self, filename: str):
        self.files_in_session.add(filename)

    def will_skip_file(self, filename: str):
        self.excluded_files.add(filename)

    def save(self):
        persistent_session = self.session_store.new_session(self.session_name)
        persistent_session.save_with_metadata(self.event_buffer, self.files_in_session, self.excluded_files)












from pycrunch_trace.client.api.trace import Trace

yoba = Trace()
yoba.start(session_name='timing_datetime')

def f():
    import time
    from datetime import datetime
    from time import sleep
    diff = 0
    for _ in range(10000):
        start_time = datetime.now()
        # sleep(0.01)
        end_time = datetime.now()
        diff += (end_time - start_time).total_seconds()
    print(diff)

    diff = 0
    for _ in range(10000):
        start_time = time.perf_counter()
        # sleep(0.01)
        end_time = time.perf_counter()
        diff += (end_time - start_time)
    print(diff)

f()
yoba.stop()
# print(yoba._tracer.simulation.simulated_code())


class GlebTracer:
    def __init__(self):
        self.current_filename = None
        pass

    def trace_calls_and_returns(self, frame, event, arg):
        filename_to_log = ''
        if self.current_filename != frame.f_code.co_filename:
            self.current_filename = frame.f_code.co_filename
            filename_to_log = self.current_filename

        print(f': {filename_to_log}:{frame.f_lineno} {event}')

        co = frame.f_code
        func_name = co.co_name
        if func_name == 'write':
            # Ignore write() calls from printing
            return
        line_no = frame.f_lineno
        filename = co.co_filename
        blacklisted = self.is_blacklisted(filename)
        if blacklisted:
            # Ignore calls not in this module
            return
        if event == 'call':
            print('* Call to function `{}` located on line {} of {}'.format(
                func_name, line_no, filename))
            return self.trace_calls_and_returns
        elif event == 'return':
            print('* return:event -> function `{}` => return value: `{}`'.format(func_name, arg))
        return

    def is_blacklisted(self, filename):
        # endswith = filename.endswith('sys_settrace_return.py')
        return False


import sys


def trace_calls(frame, event, arg):
    if event != 'call':
        return
    co = frame.f_code
    func_name = co.co_name
    if func_name == 'write':
        # Ignore write() calls from printing
        return
    func_line_no = frame.f_lineno
    func_filename = co.co_filename
    if not func_filename.endswith('sys_settrace_call.py'):
        # Ignore calls not in this module
        return
    caller = frame.f_back
    caller_line_no = caller.f_lineno
    caller_filename = caller.f_code.co_filename
    print('* Call to', func_name)
    print('*  on line {} of {}'.format(
        func_line_no, func_filename))
    print('*  from line {} of {}'.format(
        caller_line_no, caller_filename))
    return


def b():
    print('inside b()\n')


def a():
    print('inside a()\n')
    b()


sys.settrace(trace_calls)
a()


from typing import Union, Dict

from pycrunch_trace.events.base_event import Event
from pycrunch_trace.proto import message_pb2


class EventBufferInProtobuf:
    files: Dict[str, int]

    def __init__(self, event_buffer, files: Dict[str, int]):
        self.files = files
        self.event_buffer = event_buffer

    def as_bytes(self):
        session = message_pb2.TraceSession()
        for e in self.event_buffer:
            evt = self.pb_event_from_py(e)

            session.events.append(evt)

            frame = message_pb2.StackFrame()
            if e.stack:
                frame.id = e.stack.id


                if not e.stack.line:
                    frame.line = -1
                    frame.file = ''
                    frame.function_name = '??'
                    print('suka')
                    print(e.stack)
                else:
                    frame.line = e.stack.line
                    frame.file = e.stack.file
                    frame.function_name = e.stack.function_name

                if e.stack and e.stack.parent:
                    frame.parent_id = e.stack.parent.id
                else:
                    frame.parent_id = -1
            else:
                frame.id = 0

            session.stack_frames.append(frame)

        self.add_files_to_pb_envelope(session)

        return session.SerializeToString()

    def add_files_to_pb_envelope(self, session):
        for (filename, file_id) in self.files.items():
            current_file = message_pb2.File()
            # print(f'file_id: {file_id}')
            # print(f'filename: {filename}')
            current_file.id = file_id
            current_file.file = filename
            session.files.append(current_file)

    def pb_event_from_py(self, e: Event):
        evt = message_pb2.TraceEvent()
        evt.event_name = e.event_name
        evt.ts = e.ts
        if e.event_name == 'line' or e.event_name == 'method_exit':
            for key, value in e.locals.variables.items():
                pb_var = message_pb2.Variable()
                pb_var.name = key
                pb_var.value = str(value)
                evt.locals.variables.append(pb_var)

        if e.event_name == 'method_enter':
            for key, value in e.input_variables.variables.items():
                pb_var = message_pb2.Variable()
                pb_var.name = key
                pb_var.value = str(value)
                evt.input_variables.variables.append(pb_var)

        if e.event_name == 'method_exit':
            for key, value in e.return_variables.variables.items():
                pb_var = message_pb2.Variable()
                pb_var.name = key
                pb_var.value = str(value)
                evt.return_variables.variables.append(pb_var)

        evt.cursor.file = e.cursor.file
        evt.cursor.line = e.cursor.line
        evt.cursor.function_name = e.cursor.function_name
        # ptask.task.MergeFrom(task)
        if e.stack:
            evt.stack_id = e.stack.id
        return evt



import struct

from pycrunch_trace.events.method_enter import ExecutionCursor
from pycrunch_trace.tracing.simple_tracer import CallStack


def test_simple():
    x = struct.calcsize("i")
    print('size='+str(x))
    sut = CallStack()
    sut.enter_frame(ExecutionCursor('test', 1))
    sut.enter_frame(ExecutionCursor('test', 2))
    print(sut.top_level_frame_as_clone())
    sut.enter_frame(ExecutionCursor('test', 3))
    print(sut.top_level_frame_as_clone())



from typing import Dict, Any

from pycrunch_trace.file_system.human_readable_size import HumanReadableByteSize


class SessionStats:
    total_bytes: int
    total_events: int
    session_id: str
    total_samples: int


    def __init__(self, session_id: str):
        self.session_id = session_id
        self.total_events = 0
        self.total_bytes = 0
        self.total_samples = 0

    def sample(self, events, buffer_size):
        size_per_event = buffer_size / events
        print(f'{self.session_id}:{self.total_samples} avg bytes: {HumanReadableByteSize(size_per_event)} ')
        self.total_events += events
        self.total_bytes += buffer_size
        self.total_samples += 1

    def average_event_size(self):
        if self.total_events == 0:
            return 0
        return self.total_bytes / self.total_events

    def __str__(self):
        return f'SessionStats({self.session_id}) average: {self.average_event_size()}; total: {self.total_events}; bytes:{self.total_bytes}'

class PerformanceInsights:

    sessions: Dict[str, SessionStats]

    def __init__(self):
        self.sessions = dict()
        pass

    def sample(self, session_id, number_of_events, buffer_size):
        stats = self.get_or_create_session(session_id)

        stats.sample(number_of_events, buffer_size)

    def get_or_create_session(self, session_id):
        if session_id not in self.sessions:
            stats = SessionStats(session_id)
            self.sessions[session_id] = stats
        else:
            stats = self.sessions[session_id]
        return stats

    def average_event_size(self, session_id):
        stats = self.get_or_create_session(session_id)
        return stats.average_event_size()


performance_insights = PerformanceInsights()



from typing import List, Any, Optional

from pycrunch_trace.tracing.simulation import models
from pycrunch_trace.tracing.simulation.models import Frame


class SimulationEvent:
    frame: Frame
    event: str
    arg: Optional[Any]

    def __init__(self, frame: models.Frame, event: str, arg: Optional[Any]):
        self.arg = arg
        self.event = event
        self.frame = frame


class SimulationEventAsCode:
    origin: SimulationEvent

    def __init__(self, origin: SimulationEvent):
        self.origin = origin

    def as_python_code(self, index: int):
        x = f"""
def create_event_{index}():
    code_clone = models.Code()
    code_clone.co_name = '{self.origin.frame.f_code.co_name}'
    code_clone.co_filename = '{self.origin.frame.f_code.co_filename}'
    code_clone.co_argcount = {self.origin.frame.f_code.co_argcount}
    
    sim_frame = models.Frame()
    sim_frame.f_lineno = {self.origin.frame.f_lineno}
    sim_frame.f_locals = {self.origin.frame.f_locals}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, '{self.origin.event}', {self.origin.arg})

    return evt
    
"""
        return x

class DisabledSimulatorSink(object):
    def save_for_simulator(self, frame: models.Frame, event: str, arg):
        pass

class SimulatorSink:
    current_index: int
    simulation: List[str]

    def __init__(self):
        self.simulation = []
        self.current_index = 1

    def simulated_code(self):
        imports_code = """
from pycrunch_trace.simulation import models
from pycrunch_trace.tracing.simulator_sink import SimulationEvent

        """

        events = ''.join(self.simulation)

        test_code = """
def test_simulated():
    events = []"""
        for x in range(1, self.current_index):
            test_code += f'\n    events.append(create_event_{x}())'
        return imports_code + events + test_code

    def save_for_simulator(self, frame: models.Frame, event: str, arg):
        code = frame.f_code

        code_clone = models.Code()
        code_clone.co_name = code.co_name
        code_clone.co_filename = code.co_filename
        code_clone.co_argcount = code.co_argcount

        clone = models.Frame()
        clone.f_lineno = frame.f_lineno
        clone.f_locals = frame.f_locals
        clone.f_code = code_clone
        self.simulation.append(SimulationEventAsCode(SimulationEvent(clone, event, arg)).as_python_code(self.current_index))
        self.current_index += 1



from typing import Dict, List


class AbstractNetworkCommand:
    command_name: str


class StartCommand(AbstractNetworkCommand):
    def __init__(self, session_id: str):
        self.session_id = session_id
        self.command_name = 'StartCommand'

class EventsSlice(AbstractNetworkCommand):
    files: Dict[str, int]

    def __init__(self, session_id: str, chunk_number: int, events: list, files: Dict[str, int]):
        self.files = files
        self.chunk_number = chunk_number
        self.events = events
        self.command_name = 'EventsSlice'
        self.session_id = session_id

class FileContentSlice(AbstractNetworkCommand):
    files: Dict[str, int]

    def __init__(self, session_id: str, files: Dict[str, int]):
        self.command_name = 'FileContentSlice'
        self.files = files
        self.session_id = session_id


class StopCommand(AbstractNetworkCommand):
    def __init__(self, session_id: str, files_included: List[str], files_excluded: List[str]):
        self.command_name = 'StopCommand'
        self.session_id = session_id
        self.files_included = files_included
        self.files_excluded = files_excluded




from functools import wraps
from pycrunch_trace.client.api.trace import Trace



def trace(session_name_or_func=None, *decorator_args, **decorator_kws):
    """
       Starts tracing using PyCrunch tracing toolkit
   """
    def _decorator(func):
        @wraps(func)
        def wrapper(*args, **kws):
            if 'session_name_or_func' not in locals() \
                    or callable(session_name_or_func) \
                    or session_name_or_func is None:
                session_name = str(func.__name__)
            else:
                session_name = session_name_or_func
            pycrunch_tracer = Trace()
            pycrunch_tracer.start(session_name)
            try:
                results = func(*args, **kws)
                return results
            except Exception as e:
                message_code = str(e)
                raise
            finally:
                pycrunch_tracer.stop()

        return wrapper

    return _decorator(session_name_or_func) if callable(session_name_or_func) else _decorator




# TLV format:
# TAG | LENGTH | VALUE
# Header contains nested TLV triplets

TRACE_TAG_HEADER = 1

TRACE_TAG_EVENTS = 2
TRACE_TAG_FILES = 3
TRACE_TAG_METADATA = 4

HEADER_TAG_VERSION = 1
HEADER_TAG_FILES = 2
HEADER_TAG_METADATA = 3



import io


class WriteableFile:
    def __init__(self, file_name: str, buffer: bytes):
        self.buffer = buffer
        self.file_name = file_name

    def save(self):
        with io.FileIO(self.file_name, 'w') as file:
            file.write(self.buffer)



from typing import Any


class EventKeys:
    call = 'call'
    line = 'line'
    event_return = 'return'
    exception = 'exception'


class Code:
    co_filename: str

    """ function name """
    co_name: str

    """ Number of arguments (excluding keyword, *, and **) """
    co_argcount: int

    """ machine bytecode """
    co_code: bytearray

class Frame:
    f_locals: dict
    f_back: Any
    f_builtins: Any
    f_code: Code
    f_lineno: int



import io
import json
import pickle
from pathlib import Path


class Snapshot:
    def __init__(self):
        self.snapshot_directory = Path.joinpath(Path(__file__).parent, 'snapshots')

    def save(self, name: str, data):
        self.ensure_snapshot_folder_created()
        file_to_save = self.filename_for_snapshot(name)
        with io.FileIO(file_to_save, mode='w') as file:
            result = self.get_snapshot_bytes(data)
            file.write(result)

    def filename_for_snapshot(self, name):
        return self.snapshot_directory.joinpath(f'{name}.pycrunch-trace')

    def load(self, name: str):
        self.ensure_snapshot_folder_created()
        file_to_load = self.filename_for_snapshot(name)
        with io.FileIO(file_to_load, mode='r') as file:
            buffer = file.readall()
            # try:
                # result = json.loads(buffer)
            # except:
            result = pickle.loads(buffer)
            return result

    def get_snapshot_bytes(self, data):
        # if isinstance(data, dict):
        #     return json.dumps(data, indent=2).encode('utf-8')
        return pickle.dumps(data)


    def ensure_snapshot_folder_created(self):
        if not self.snapshot_directory.exists():
            self.snapshot_directory.mkdir(exist_ok=True)

snapshot = Snapshot()


from pycrunch.insights import trace

from pycrunch_trace.events.method_enter import ExecutionCursor
from pycrunch_trace.tracing.simple_tracer import CallStack


def test_1():
    sut = CallStack()
    sut.enter_frame(ExecutionCursor('a', 1))
    sut.enter_frame(ExecutionCursor('b', 2))
    sut.enter_frame(ExecutionCursor('c', 3))
    sut.new_cursor_in_current_frame(ExecutionCursor('b', 3))
    sut.exit_frame()

    x = sut.current_frame()
    ex = sut.top_level_frame_as_clone()
    # trace(yyy=str(x))
    trace(ex=str(ex))



from datetime import datetime


def some_method(some_number):
    a = 2
    b = 3
    b = 3*4
    b = 3*4
    b = 1*4
    result = a + b + some_number
    return result


def another_m(some_new_var):
    x = 2
    for i in range(some_new_var):
        if i % 2 == 0:
            x = 3 + i
        else:
            x = 2 * i

def string_m():
    return str(datetime.now())
# x = some_method(1)
# print(x)



from pycrunch_trace.file_system.session_store import SessionStore
from pycrunch_trace.file_system.trace_session import TraceSession
from pycrunch_trace.tests.test_buffer import build_testing_events

def a():
    pass

def test_1():
    x = TraceSession()
    x.save('a', build_testing_events())

def test_session_store():
    x = SessionStore()
    session = x.load_session('a')
    session.load_metadata()
    print(vars(session.metadata))
def test_load_sessions():
    x = SessionStore()
    print(x.all_sessions())

def test_multiple_files():
    x = TraceSession('aaa')
    x.did_enter_traceable_file('a')
    x.did_enter_traceable_file('a2')
    x.did_enter_traceable_file('b')
    print('aaa')
    x.will_skip_file('c')
    x.will_skip_file('c')
    x.will_skip_file('d')
    x.buffer_became_available(build_testing_events())
    x.save()



import io
import shutil
import struct
from datetime import datetime
from pathlib import Path
from typing import Dict, List

import jsonpickle

from pycrunch_trace.file_system import tags
from pycrunch_trace.file_system.human_readable_size import HumanReadableByteSize
from pycrunch_trace.file_system.persisted_session import PersistedSession, TraceSessionMetadata
from pycrunch_trace.file_system.session_store import SessionStore
from pycrunch_trace.file_system.trace_file import TraceFile
from pycrunch_trace.server.chunks_ordering import PyCrunchTraceException
from pycrunch_trace.server.incoming_traces import incoming_traces


class TracePersistence:
    trace_files: Dict[str, TraceFile]

    def __init__(self):
        x = SessionStore()
        x.ensure_recording_directory_created()
        self.rec_dir = x.recording_directory
        self.trace_files = dict()
        pass

    def initialize_file(self, session_id: str):
        dir_path = Path(self.rec_dir)
        session_id = session_id
        rec_dir = dir_path.joinpath(session_id)
        if rec_dir.exists():
            shutil.rmtree(rec_dir)

        tf = TraceFile(session_id, self.get_chunked_trace_output_file(session_id))
        self.trace_files[session_id] = tf
        tf.write_header_placeholder()

    def recording_complete(self, session_id, files_included: List[str], files_excluded: List[str]):
        metadata_bytes, metadata_file_path = self.get_metadata_bytes(session_id, files_included, files_excluded)
        self.update_file_header_metadata_section(session_id, len(metadata_bytes))
        self.flush_chunk(session_id, tags.TRACE_TAG_METADATA, metadata_bytes)

        with io.FileIO(metadata_file_path, mode='w') as file:
            bytes_written = file.write(metadata_bytes)
            print(f'metadata saved to {metadata_file_path}')


    def get_metadata_bytes(self, session_id, files_included: List[str], files_excluded: List[str]):
        dir_path = Path(self.rec_dir)
        rec_dir = dir_path.joinpath(session_id)
        x = SessionStore()
        target_chunk_file = rec_dir.joinpath(PersistedSession.chunked_recording_filename)
        bytes_written = target_chunk_file.stat().st_size
        metadata_file_path = rec_dir.joinpath(PersistedSession.metadata_filename)
        meta = TraceSessionMetadata()
        meta.files_in_session = files_included
        meta.excluded_files = files_excluded
        meta.file_size_in_bytes = bytes_written
        meta.file_size_on_disk = str(HumanReadableByteSize(bytes_written))
        trace_in_progress = incoming_traces.get_session_with_id(session_id)
        meta.events_in_session = trace_in_progress.total_events
        meta.start_time = trace_in_progress.started_at
        meta.end_time = datetime.utcnow()
        meta.name = str(session_id)
        result = jsonpickle.dumps(meta, unpicklable=False)
        metadata_bytes = result.encode('utf-8')
        return metadata_bytes, metadata_file_path

    def flush_chunk(self, session_id, tag_id: int, bytes_to_write):
        trace_file = self._get_trace_file_or_throw(session_id)
        trace_file.flush_chunk(tag_id, bytes_to_write)

    def get_chunked_trace_output_file(self, session_id):
        dir_path = Path(self.rec_dir)
        rec_dir = dir_path.joinpath(session_id)
        x = SessionStore()
        x.ensure_directory_created(rec_dir)
        target_file = rec_dir.joinpath(PersistedSession.chunked_recording_filename)
        return target_file

    def get_write_mode_if_file_exist(self, target_file):
        target_mode = 'a'
        if not target_file.exists():
            target_mode = 'w'
        return target_mode

    def update_file_header_files_section(self, session_id, total_bytes):
        trace_file = self._get_trace_file_or_throw(session_id)
        trace_file.update_file_header_files_section(total_bytes)


    def _get_trace_file_or_throw(self, session_id) -> TraceFile:
        trace_file = self.trace_files.get(session_id)
        if not trace_file:
            raise PyCrunchTraceException(f'Cannot find trace file with session id {session_id}')
        return trace_file

    def update_file_header_metadata_section(self, session_id, metadata_bytes_len):
        trace_file = self._get_trace_file_or_throw(session_id)
        trace_file.update_file_header_metadata_section(metadata_bytes_len)




from typing import List

from pycrunch_trace.client.networking.commands import EventsSlice, FileContentSlice


class AbstractRecordingStrategy:
    def prepare(self):
        pass

    def recording_start(self, session_id: str):
        pass

    def recording_stop(self, session_id: str, files_included: List[str], files_excluded: List[str]):
        pass

    def recording_slice(self, x: EventsSlice):
        pass

    def files_slice(self, x: FileContentSlice):
        pass

    def clean(self):
        pass


import io
import json
import pickle
from datetime import datetime
from pathlib import Path
from typing import List

import jsonpickle

from pycrunch_trace.events.event_buffer_in_protobuf import EventBufferInProtobuf
from pycrunch_trace.file_system.chunked_trace import ChunkedTrace
from pycrunch_trace.file_system.human_readable_size import HumanReadableByteSize
from pycrunch_trace.proto import message_pb2
import logging

logger = logging.getLogger(__name__)


class TraceSessionMetadata:
    # time in UTC
    start_time: datetime
    end_time: datetime
    # size in bytes
    file_size_in_bytes: int
    file_size_on_disk: str
    files_in_session: List[str]

    events_in_session: int

    working_directory: str

    name: str


class LazyLoadedSession:
    metadata: TraceSessionMetadata
    trace_filename: Path
    metadata_file: Path
    raw_metadata: dict

    def __init__(self, buffer_file: Path, metadata_file: Path, chunked: bool):
        self.chunked = chunked
        self.trace_filename = buffer_file
        self.metadata_file = metadata_file
        self.raw_metadata = None

    def load_buffer(self):
        if not self.chunked:
            return self.load_unchunked()
        else:
            return ChunkedTrace(self.trace_filename).events()

    def load_unchunked(self):
        with io.FileIO(self.trace_filename, mode='r') as file:
            buffer = file.readall()
            # try:
            # result = json.loads(buffer)
            # except:

            # result = pickle.loads(buffer)

            result = message_pb2.TraceSession()
            result.ParseFromString(buffer)
            return result

    def load_metadata(self):
        with io.FileIO(self.metadata_file, 'r') as file:
            file_bytes = file.readall()
            json_representation = file_bytes.decode('utf-8')
            json_dict = json.loads(json_representation)
            self.raw_metadata = json_dict
            meta = TraceSessionMetadata()
            meta.files_in_session = json_dict.get('files_in_session')
            meta.excluded_files = json_dict.get('events_in_session')
            meta.events_in_session = json_dict.get('events_in_session')
            meta.file_size_in_bytes = json_dict.get('file_size_in_bytes')
            meta.file_size_on_disk = json_dict.get('file_size_on_disk')
            meta.name = json_dict.get('name')
            self.metadata = meta


class PersistedSession:
    def __init__(self, session_directory: Path):
        self.session_directory = session_directory

    metadata_filename = 'pycrunch-trace.meta.json'
    recording_filename = 'session.pycrunch-trace'
    chunked_recording_filename = 'session.chunked.pycrunch-trace'

    pass

    def save_with_metadata(self, event_buffer, files_in_session, excluded_files):
        file_to_save = self.session_directory.joinpath(self.recording_filename)
        bytes_written = -42
        with io.FileIO(file_to_save, mode='w') as file:
            try:
                result = self.serialize_to_bytes(event_buffer)
                bytes_written = file.write(result)
            except Exception as ex:
                logger.exception('failed to save session', exc_info=ex)

        meta = TraceSessionMetadata()
        meta.files_in_session = list(files_in_session)
        meta.excluded_files = list(excluded_files)
        meta.file_size_in_bytes = bytes_written
        meta.file_size_on_disk = str(HumanReadableByteSize(bytes_written))
        meta.events_in_session = len(event_buffer)
        meta.name = str(self.session_directory)
        print(f'tracing --- protobuf binary array results saved to file {file_to_save}')

        self.save_metadata(self.session_directory, meta)

    def serialize_to_bytes(self, event_buffer):
        return EventBufferInProtobuf(event_buffer, x.file_map).as_bytes()
        # todo add multiple serialization plugin/options
        # return pickle.dumps(event_buffer)

    def save_metadata(self, session_directory: Path, meta: TraceSessionMetadata):
        metadata_file_path = session_directory.joinpath(self.metadata_filename)

        with io.FileIO(metadata_file_path, mode='w') as file:
            result = self.serialize_to_json(meta)
            bytes_written = file.write(result.encode('utf-8'))

    def serialize_to_json(self, meta) -> str:
        return jsonpickle.dumps(meta, unpicklable=False)

    @classmethod
    def load_from_directory(cls, load_from_directory: Path) -> LazyLoadedSession:
        chunked = False
        joinpath = load_from_directory.joinpath(PersistedSession.recording_filename)
        if not joinpath.exists():
            joinpath = load_from_directory.joinpath(PersistedSession.chunked_recording_filename)
            chunked = True
        print(joinpath)
        return LazyLoadedSession(joinpath, load_from_directory.joinpath(PersistedSession.metadata_filename), chunked)



from pycrunch_trace.client.api.trace import Trace


def c(number):
    print('in c')
    response = number * 2
    return response


def b(y):
    print('opa')
    return y * 2


def a(x):
    y = x + 1
    res = b(y)
    y = res + 1
    res = c(res + 1)
    return res

yoba = Trace()
yoba.start('multiple_stack')

a(2)

yoba.stop()


from queue import Queue
from typing import Dict, Any


class PyCrunchTraceException(Exception):
    pass


class PyCrunchTraceServerException(PyCrunchTraceException):
    pass


class ReceivedChunks:
    def __init__(self, session_id: str):
        self.session_id = session_id
        self.received_chunks = set()

    def did_receive_chunk(self, chunk_number):
        self.throw_if_chunk_out_of_order(chunk_number)
        self.received_chunks.add(chunk_number)

    def throw_if_chunk_out_of_order(self, chunk_number):
        self.throw_if_chunk_already_received(chunk_number)
        if len(self.received_chunks) == 0:
            self.throw_if_first_chunk_lost(chunk_number)
        else:
            self.throw_if_previous_chunk_lost(chunk_number)

    def throw_if_previous_chunk_lost(self, chunk_number):
        current_max = max(self.received_chunks)
        if chunk_number - 1 > current_max:
            raise PyCrunchTraceServerException(f'{self.session_id}: received {chunk_number} but {current_max} is expected for')

    def throw_if_first_chunk_lost(self, chunk_number):
            if chunk_number != 1:
                raise PyCrunchTraceServerException(f'{self.session_id}: received {chunk_number} but expected chunk #1')

    def throw_if_chunk_already_received(self, chunk_number):
        if chunk_number in self.received_chunks:
            raise PyCrunchTraceServerException(f'already received {chunk_number} for {self.session_id}')


class ChunksOrdering:
    order_by_session: Dict[str, ReceivedChunks]

    def __init__(self):
        self.order_by_session = dict()
        self.history = Queue(maxsize=200)

    def session_will_start(self, session_id):
        self.order_by_session[session_id] = ReceivedChunks(session_id)

    def did_receive_chunk(self, session_id, chunk_number):
        if session_id not in self.order_by_session:
            raise PyCrunchTraceServerException(f'{session_id}: not found')

        self.order_by_session[session_id].did_receive_chunk(chunk_number)

    def session_will_finish(self, session_id):
        self.history.put_nowait(self.order_by_session[session_id])
        del self.order_by_session[session_id]


chunks_ordering: ChunksOrdering = ChunksOrdering()



from pathlib import Path
from typing import List, Any

import yaml

from ..filters import CustomFileFilter
from ..oop import Directory, AbstractFile
from ..oop import File


class Profiles:
    def __init__(self, directory: Directory):
        self.directory = directory


def test_profiles():
    directory = Directory(Path.joinpath(Path(__file__).parent.parent, 'pycrunch-profiles'))
    print(directory)
    all = directory.folders()
    all = directory.files('yaml')
    print(str(all[0].short_name()))




def test_profile_should_have_excluded():
    buffer = b"""
exclusions:
  - no_trace.py
  - /Library/
    """
    sut = CustomFileFilter(File.Mock(buffer))

    assert not sut.should_trace('no_trace.py')
    assert not sut.should_trace('/Library/a.py')
    assert sut.should_trace('b.py')
    assert sut.should_trace('/code/b.py')


def test_profile_trace_vars_false():
    buffer = b"""
trace_variables: false
    """
    sut = CustomFileFilter(File.Mock(buffer))

    assert not sut.should_record_variables()


def profiles_dir():
    return Directory(Path.joinpath(Path(__file__).parent, 'pycrunch-profiles'))


import pycrunch_trace.events.method_enter as events
import collections


class CallStack:
    def __init__(self):
        self.stack = collections.deque()

    def enter_frame(self, execution_cursor: events.ExecutionCursor):
        parent_frame = self.get_parent_frame()
        # print(f"{execution_cursor.file}:{execution_cursor.line} -> {parent_frame} ")
        frame = events.StackFrame.new(parent_frame, execution_cursor)
        self.stack.append(frame)

    def get_parent_frame(self):
        if len(self.stack) > 0:
            return self.stack[-1]
        return None

    def new_cursor_in_current_frame(self, new_cursor: events.ExecutionCursor):
        stack_frame: events.StackFrame = self.top_level_frame_as_clone()
        stack_frame.line = new_cursor.line
        stack_frame.file = new_cursor.file
        stack_frame.function_name = new_cursor.function_name
        # todo this is probably dirty hack?
        # or just replacing last-known stack frame
        #  todo what about performance ?
        if len(self.stack) > 0:
            self.stack.pop()
            self.stack.append(stack_frame)
            # self.stack[-1] = stack_frame
        else:
            # session just begin, not yet in any stack
            self.stack.append(stack_frame)


    def exit_frame(self):
        self.stack.pop()

    def top_level_frame_as_clone(self):
        current: events.StackFrame = self.current_frame()
        # print('top_level_frame_as_clone ->' + str(current))
        return events.StackFrame.clone(current)
        # ???
        # return current

    def current_frame(self):
        frame = self.get_parent_frame()
        return frame



import socketio

from . import version


class TracingClient:

    def __init__(self, host_url: str):
        self.sio = socketio.Client()
        connection_headers = dict(
            version=version.version,
            product='pycrunch-tracing-node',
        )
        self.sio.connect(url=host_url, headers=connection_headers, transports=['websocket'])

        @self.sio.event
        def message(data):
            print('CLIENT: I received a message!')

        @self.sio.on('my message')
        def on_message(data):
            print('CLIENT: I received a message!')

        @self.sio.event
        def connect():
            print("CLIENT: I'm connected!")

        @self.sio.event
        def connect_error():
            print("CLIENT: The connection failed!")

        @self.sio.event
        def disconnect():
            print("CLIENT: I'm disconnected!")

    def push_message(self, entire_tracing_sesssion):
        # dumps = pickle.dumps(entire_tracing_sesssion)
        # print(f'dumped {len(dumps)} bytes')
        try:
            print(f' ...sending bytes')

            self.sio.emit('event', dict(
                action='new_recording',
                # buffer=dumps,
            ))
            print(f' ...sent')
        except Exception as e:
            print('  -- !fail to send')
            print(str(e))

    def disconnect(self):

        self.sio.disconnect()



from collections import deque
from typing import List

from pycrunch_trace.events.base_event import Event


class ArrayCommandBuffer:
    def __init__(self):
        # todo swap implementation to linked list?
        self._command_buffer = []

    def add_event(self, evt):
        self._command_buffer.append(evt)

    def finish_chunk(self) -> List[Event]:
        old_buffer = self._command_buffer
        self._command_buffer = []
        return old_buffer

    def how_many_events(self):
        return len(self._command_buffer)

class DequeCommandBuffer:
    def __init__(self):
        self._command_buffer = deque()

    def add_event(self, evt):
        self._command_buffer.append(evt)

    def finish_chunk(self) -> List[Event]:
        old_buffer = self._command_buffer.copy()
        self._command_buffer.clear()
        return list(old_buffer)

    def how_many_events(self):
        return len(self._command_buffer)


class HumanReadableByteSize:
    def __init__(self, total_bytes: int):
        self.total_bytes = total_bytes

    def __str__(self):
        return self.human_readable_size(self.total_bytes)

    @staticmethod
    def human_readable_size(size, decimal_places=1):
        if size < 1024:
            return f"{round(size)} bytes"
        for unit in ['B', 'Kb', 'MB', 'GB', 'TB']:
            if size < 1024.0:
                break
            size /= 1024.0
        return f"{size:.{decimal_places}f} {unit}"



import pytest

from pycrunch_trace.server.chunks_ordering import ChunksOrdering, PyCrunchTraceServerException


def test_add_chunk_works():
    sut = ChunksOrdering()
    sut.session_will_start(session_id='s1')
    sut.did_receive_chunk(session_id='s1', chunk_number=1)

    assert 1 in sut.order_by_session['s1'].received_chunks

def test_duplicate_chunk_raises():
    with pytest.raises(PyCrunchTraceServerException):
        sut = ChunksOrdering()
        sut.session_will_start(session_id='s1')
        sut.did_receive_chunk(session_id='s1', chunk_number=1)
        sut.did_receive_chunk(session_id='s1', chunk_number=2)
        sut.did_receive_chunk(session_id='s1', chunk_number=2)

def test_lost_chunk_raises():
    with pytest.raises(PyCrunchTraceServerException):
        sut = ChunksOrdering()
        sut.session_will_start(session_id='s1')
        sut.did_receive_chunk(session_id='s1', chunk_number=1)
        sut.did_receive_chunk(session_id='s1', chunk_number=2)
        # lost 6 frames
        sut.did_receive_chunk(session_id='s1', chunk_number=8)

def test_first_chunk_lost():
    with pytest.raises(PyCrunchTraceServerException):
        sut = ChunksOrdering()
        sut.session_will_start(session_id='s1')
        sut.did_receive_chunk(session_id='s1', chunk_number=2)

def test_happy_case_no_errors():
    sut = ChunksOrdering()
    sut.session_will_start(session_id='s1')
    sut.did_receive_chunk(session_id='s1', chunk_number=1)
    sut.did_receive_chunk(session_id='s1', chunk_number=2)
    sut.did_receive_chunk(session_id='s1', chunk_number=3)


def test_multiple_sessions():
    sut = ChunksOrdering()
    sut.session_will_start(session_id='s1')
    sut.did_receive_chunk(session_id='s1', chunk_number=1)
    sut.session_will_start(session_id='s2')
    sut.did_receive_chunk(session_id='s2', chunk_number=1)
    sut.did_receive_chunk(session_id='s2', chunk_number=2)
    sut.session_will_start(session_id='s3')
    sut.did_receive_chunk(session_id='s3', chunk_number=1)
    sut.did_receive_chunk(session_id='s3', chunk_number=2)


def test_chunk_after_session_finished():
    with pytest.raises(PyCrunchTraceServerException):
        sut = ChunksOrdering()
        sut.session_will_start('s1')
        sut.did_receive_chunk('s1', 1)
        sut.session_will_finish('s1')
        sut.did_receive_chunk('s1', 1)



from pycrunch_trace.events.method_enter import MethodEnterEvent, ExecutionCursor, StackFrame, MethodExitEvent
from pycrunch_trace.tracing.file_map import FileMap


def test_file_id():
    sut = FileMap()
    assert 1 == sut.file_id('a.py')


def test_same_file_should_return_same_id():
    sut = FileMap()
    assert 1 == sut.file_id('a.py')
    assert 1 == sut.file_id('a.py')
    assert 1 == sut.file_id('a.py')

def test_multiple_files():
    sut = FileMap()
    assert 1 == sut.file_id('a.py')
    assert 2 == sut.file_id('b.py')
    assert 3 == sut.file_id('c.py')
    assert 1 == sut.file_id('a.py')
    assert 2 == sut.file_id('b.py')
    assert 3 == sut.file_id('c.py')

def test_a():
    from pycrunch_trace.client.networking.client_trace_introspection import ClientTraceIntrospection
    sut = ClientTraceIntrospection()
    files = dict()
    files['c'] = 7
    files['a'] = 1
    files['b'] = 2
    sut.save_events([
        MethodEnterEvent(
            ExecutionCursor(7,1,'1'),
            StackFrame(-1, 1,1,'a'),
            1),
        MethodExitEvent(
            ExecutionCursor(1, 1, '1'),
            StackFrame(-1, 1, 1, 'a'),
            2),
        MethodExitEvent(
            ExecutionCursor(1, 1, '1'),
            StackFrame(-1, 1, 1, 'a'),
            2),
    ])
    sut.print_to_console(files)


import pycrunch_trace.events.method_enter as e
from pycrunch_trace.config import config
from pycrunch_trace.serialization import to_string


def test_simple():
    command_stack = build_testing_events()
    # print(command_stack)
    print(to_string(command_stack))





def build_testing_events():
    command_stack = []
    stack = ['a:11', 'b:12']
    method_enter_event = e.MethodEnterEvent(e.ExecutionCursor(config.absolute_path, 2), stack)
    method_enter_event.input_variables.push_variable('some_number', 1)
    command_stack.append(method_enter_event)
    line_1 = e.LineExecutionEvent(e.ExecutionCursor(config.absolute_path, 3), stack)
    line_1.locals.push_variable('a', 2)
    command_stack.append(line_1)
    line_2 = e.LineExecutionEvent(e.ExecutionCursor(config.absolute_path, 4), stack)
    line_2.locals.push_variable('a', 2)
    line_2.locals.push_variable('b', 3)
    command_stack.append(line_2)
    line_3 = e.LineExecutionEvent(e.ExecutionCursor(config.absolute_path, 5), stack)
    line_3.locals.push_variable('a', 2)
    line_3.locals.push_variable('b', 3)
    line_3.locals.push_variable('result', 6)
    command_stack.append(line_3)
    line_return = e.MethodExitEvent(e.ExecutionCursor(config.absolute_path, 6), stack)
    line_return.return_variables.push_variable('return_value', 6)
    command_stack.append(line_return)
    return command_stack



class SafeFilename:
    possibly_unsafe_filename: str

    def __init__(self, possibly_unsafe_filename: str):
        self.possibly_unsafe_filename = possibly_unsafe_filename

    def __str__(self):
        return self._make_safe_filename()

    def _make_safe_filename(self):
        def safe_char(c):
            if c.isalnum():
                return c
            else:
                return "_"

        return "".join(safe_char(c) for c in self.possibly_unsafe_filename).rstrip("_")





import time


class Clock:
    """
    Counts time since creating of the object
    """
    def __init__(self):
        self.started_at = self._system_clock()
        pass

    def now(self):
        return (time.perf_counter() - self.started_at) * 1000

    def _system_clock(self):
        return time.perf_counter()



allowed_types = [int, str, float, dict, type(None), bool]
# allowed_types = [int, str, float,  type(None), bool]


def can_trace_type(variable):
    # return False
    current_type = type(variable)
    if current_type in allowed_types:
        return True

    # print(current_type)

    return False


class AbstractFileFilter:
    def should_trace(self, filename: str) -> bool:
        self._abstract_exception()

    def should_record_variables(self) -> bool:
        self._abstract_exception()

    def _abstract_exception(self):
        raise Exception('AbstractFileFilter: called too abstract should_trace function')



class TracerPerf:
    def __init__(self):
        self.total_samples = 1
        self.total_time = 0.00
        pass

    def did_execute_line(self, ts_diff: float):
        self.total_samples += 1
        self.total_time += ts_diff

    def print_avg_time(self):
        each = 1
        should_print = self.total_samples % each == 0
        should_print = True
        if should_print:
            time_per_sample = self.total_time / self.total_samples
            print(f'total_samples - {self.total_samples}')
            print(f'total overhead time - {self.total_time}')
            print(f'{time_per_sample:.5f} ms avg call time overhead')


import io

import yaml

from . import shared

from pycrunch_trace.events.event_buffer_in_protobuf import EventBufferInProtobuf
from pycrunch_trace.file_system.session_store import SessionStore
from pycrunch_trace.file_system.trace_session import TraceSession
from pycrunch_trace.filters import CustomFileFilter
from pycrunch_trace.oop import Directory, WriteableFile
from pycrunch_trace.oop.file import File
from pycrunch_trace.session import active_clients
from pycrunch_trace.session.snapshot import snapshot
import pickle


import logging

from .incoming_traces import incoming_traces
from .state import connections
from ..config import config
from ..file_system.human_readable_size import HumanReadableByteSize
from ..file_system.persisted_session import TraceSessionMetadata

logger = logging.getLogger(__name__)


@shared.tracer_socket_server.event
async def connect(sid, environ):
    print("connect -", sid)
    # print("connect - environ", environ)
    product_name = environ.get('HTTP_PRODUCT')
    if product_name:
        if product_name == 'pycrunch-tracing-node':
            version = environ.get('HTTP_VERSION')

            connections.tracer_did_connect(sid, version)
            await shared.tracer_socket_server.emit('front', dict(
                event_name='new_tracker',
                sid=sid,
                version=version,
            ))


async def new_recording(req, sid):
    logger.info('Started saving new recording')
    event_buffer_bytes = req.get('buffer')
    # todo this is double loading
    if (event_buffer_bytes):
        x: TraceSession = pickle.loads(event_buffer_bytes)
        x.save()

    logger.info('Recording saved successfully')
    await load_sessions(None)
    # await sio.emit('reply', event_buffer)

total_bytes = 0



@shared.tracer_socket_server.event
async def event(sid, req):
    # print(req)
    action: str = req.get('action')
    logger.info(f'WebSocket event: {action}')

    if action == 'load_buffer':
        pass
    elif action == 'load_file':
        await load_file_event(req, sid)
    elif action == 'load_profiles':
        await load_profiles_event(req, sid)
    elif action == 'load_sessions':
        await load_sessions(sid)
    elif action == 'load_profile_details':
        await load_profile_details(req, sid)
    elif action == 'load_single_session':
        await load_single_session(req, sid)
    elif action == 'save_profile_details':
        await save_profile_details(req, sid)
    elif action == 'new_recording':
        print('new_recording')
        await new_recording(req, sid)
    else:
        await shared.tracer_socket_server.emit('reply_unknown', room=sid)


async def load_sessions(sid):
    logging.debug('Loading sessions')
    store = SessionStore()
    all_names = store.all_sessions()
    result = []
    for name in all_names:
        try:
            lazy_loaded = store.load_session(name)
            lazy_loaded.load_metadata()
            metadata = lazy_loaded.raw_metadata
            metadata['short_name'] = name
        except:
            metadata = dict()
            metadata['short_name'] = name
            metadata['events_in_session'] = 'missing meta'

        result.append(metadata)

    logging.debug(f'Sessions loaded, sending back to client {sid}')
    await shared.tracer_socket_server.emit('session_list_loaded', result, room=sid)
    pass


async def load_single_session(req, sid):
    logger.info('begin: load_single_session...')
    store = SessionStore()
    session_name = req.get('session_name')
    logging.info(f'Loading session {session_name}')
    ses = store.load_session(session_name)

    # await sio.emit('reply', to_string(buffer), room=sid)
    try:

        logger.info('sending reply...')

        file_as_bytes = ses.load_buffer().SerializeToString()
        logger.info('bytes loaded...')
        #  todo maybe send back in chunks?
        await shared.tracer_socket_server.emit('reply', data=file_as_bytes
                                               , room=sid
                                               )
        logger.info('Event sent')

    except Exception as ex:
        logger.exception('Failed to load session ' + session_name, exc_info=ex)


async def save_profile_details(req, sid):
    logger.debug(f'save_profile_details: `{req}`')
    profile = req.get('profile')
    profile_name = profile.get('profile_name')
    xxx = yaml.dump(profile)
    logger.debug(xxx)
    profiles__joinpath = config.package_directory.joinpath('pycrunch-profiles').joinpath(profile_name)

    WriteableFile(profiles__joinpath, xxx.encode('utf-8')).save()


async def load_file_event(req, sid):
    file_to_load = req.get('file_to_load')
    logger.debug(f'file_to_load: `{file_to_load}`')
    with io.open(file_to_load, 'r', encoding='utf-8') as f:
        lines = f.read()
        await shared.tracer_socket_server.emit('file_did_load', dict(filename=file_to_load, contents=lines), room=sid)


async def load_profile_details(req, sid):
    d = Directory(config.package_directory.joinpath('pycrunch-profiles'))
    profile_name = req.get('profile_name')
    joinpath = config.package_directory.joinpath('pycrunch-profiles').joinpath(profile_name)
    print(joinpath)
    fff = CustomFileFilter(File(joinpath))

    raw = fff.all_exclusions()

    await shared.tracer_socket_server.emit('profile_details_loaded', dict(
        exclusions=raw,
        trace_variables=fff.should_record_variables(),
        profile_name=profile_name), room=sid)


async def load_profiles_event(req, sid):
    d = Directory(config.package_directory.joinpath('pycrunch-profiles'))
    res = d.files('yaml')
    print(res)
    raw = []
    for f in res:
        raw.append(f.short_name())
    await shared.tracer_socket_server.emit('profiles_loaded', dict(profiles=raw), room=sid)


@shared.tracer_socket_server.event
async def disconnect(sid):
    logging.info(f'disconnect {sid}')
    if connections.tracer_did_disconnect(sid):
        logging.debug(f' -- sending notification about disconnected tracker {sid}')
        await shared.tracer_socket_server.emit('front', dict(
            event_name='tracker_did_disconnect',
            sid=sid,
        ))



from pycrunch_trace.client.api import trace
from pycrunch_trace.tracing.file_map import FileMap




@trace
def run():
    loc = dict()
    fff = FileMap()
    id__ = fff.file_id('test')
    loc['fff'] = fff
    loc['id__'] = id__
    loc['file'] = 'test'



run()


from pycrunch_trace.client.api.trace import Trace

def d___():
    a = 1
    b = 2
    c = 3

def c___():
    a = 1
    b = 2
    d___()
    c = 3


def b___():
    a = 1
    c___()
    b = 2
    c = 3



def a___():
    a = 1
    b = 2
    b___()
    c = 3




tracer = Trace()
tracer.start('recording_name')

a___()

tracer.stop()
# print(y._tracer.simulation.simulated_code())
# import pydevd_pycharm
# pydevd_pycharm.settrace('localhost', port=44441, stdoutToServer=True, stderrToServer=True)
# for (x, cmd) in enumerate(y.command_buffer):
#     print(str(x+1) + ' - line:' + str(cmd.cursor.line) + ':' + str(cmd))


import io
import struct
from pathlib import Path

from pycrunch_trace.proto import message_pb2


class ChunkedTrace:
    # size in bytes; this return 4 on my machine
    header_size = struct.calcsize("i")

    def __init__(self, filename: Path):
        self.filename = filename


    def events(self) -> list:
        file_map = dict()
        entire_session = message_pb2.TraceSession()
        events_so_far = 0
        with io.FileIO(self.filename, 'r') as file_to_read:
            while True:
                header_bytes = file_to_read.read(ChunkedTrace.header_size)
                if len(header_bytes) <= 0:
                    print(f' -- Read to end')

                    break
                next_chunk_length = struct.unpack('i', header_bytes)[0]
                print(f'next_chunk_length {next_chunk_length}')

                read_bytes = file_to_read.read(next_chunk_length)
                interrim = message_pb2.TraceSession()
                try:
                    interrim.ParseFromString(read_bytes)
                except Exception as eeeeeee:
                    print(eeeeeee)
                    raise

                for stack_frame in interrim.stack_frames:
                    entire_session.stack_frames.append(stack_frame)
                print(f'total stack_frames {len(entire_session.stack_frames)}')
                events_so_far += len(interrim.events)
                print(f'total events {events_so_far}')
                for event in interrim.events:
                    entire_session.events.append(event)

                for f in interrim.files:
                    file_map[f.file] = f.id

                if events_so_far > 10000000:
                    break
        for (file, file_id) in file_map.items():
            result_file = message_pb2.File()
            result_file.id = file_id
            result_file.file = file
            entire_session.files.append(result_file)

        return entire_session



from pycrunch_trace import config
from pycrunch_trace.filters import DefaultFileFilter


def test_included():
    sut = DefaultFileFilter()
    assert sut.should_trace(config.absolute_path)

def test_not_included():
    sut = DefaultFileFilter()
    assert not sut.should_trace('/crap/module/x.py')



import sys

import pycrunch_trace.tracing.simulation.models as models
import pycrunch_trace.events.method_enter as events
from pycrunch_trace.client.command_buffer import ArrayCommandBuffer
from pycrunch_trace.client.networking.commands import EventsSlice
from pycrunch_trace.file_system.trace_session import TraceSession
from pycrunch_trace.filters import AbstractFileFilter
from pycrunch_trace.oop import Clock
from pycrunch_trace.tracing.call_stack import CallStack
from pycrunch_trace.tracing.file_map import FileMap
from pycrunch_trace.tracing.perf import TracerPerf
from pycrunch_trace.tracing.simulation import EventKeys
from pycrunch_trace.tracing.simulator_sink import SimulatorSink, DisabledSimulatorSink


class SimpleTracer:
    file_map: FileMap
    event_number: int
    file_filter: AbstractFileFilter
    event_buffer: ArrayCommandBuffer
    call_stack: CallStack
    session: TraceSession
    simulation: SimulatorSink
    events_so_far: int

    def __init__(self, event_buffer, session_name, file_filter: AbstractFileFilter, clock: Clock, queue):
        self.event_number = 1
        self.events_so_far = 0
        self.clock = clock
        self.event_buffer = event_buffer
        self.file_filter = file_filter
        self.should_record_variables = file_filter.should_record_variables()
        self.call_stack = CallStack()
        self.session = TraceSession()
        self.simulation = DisabledSimulatorSink()
        # self.simulation = SimulatorSink()
        self.file_map = FileMap()
        self.perf = TracerPerf()
        self.queue = queue
        self.max_events_before_send = 1000
        self.threshold_before_switching_to_sampling = 1000000
        self.skip = False
        # interval = sys.getscheckinterval()
        # print(f'interval {interval}')
        # sys.setcheckinterval(100000000)

        # interval = sys.getswitchinterval()
        #
        # print(f'getswitchinterval {interval}')
        # sys.setswitchinterval(3)
        # interval = sys.getswitchinterval()
        # print(f'getswitchinterval {interval}')

    def simple_tracer(self, frame: models.Frame, event: str, arg):
        entered_at = self.clock.now()
        self.process_events(entered_at, event, frame, arg)
        # self.simulation.save_for_simulator(frame, event, arg)

        # print(f"[{co.co_argcount}]{event}: {func_name} {line_no} -> {arg}")
        # print(f"   {frame.f_locals}")

        end_at = self.clock.now()
        diff = end_at - entered_at
        self.perf.did_execute_line(diff)
        return self.simple_tracer

    def process_events(self, now: float, event: str, frame: models.Frame, arg):
        file_path_under_cursor = frame.f_code.co_filename
        if not self.file_filter.should_trace(file_path_under_cursor):
            will_record_current_event = False
            self.session.will_skip_file(file_path_under_cursor)
        else:
            will_record_current_event = True
            self.events_so_far += 1
            self.session.did_enter_traceable_file(file_path_under_cursor)

        if event == EventKeys.line:
            over_threshold = self.events_so_far > self.threshold_before_switching_to_sampling
            if will_record_current_event and not over_threshold:
                cursor = self.create_cursor(file_path_under_cursor, frame)
                self.call_stack.new_cursor_in_current_frame(cursor)

                stack = self.get_execution_stack()
                current = events.LineExecutionEvent(cursor, stack, now)
                if self.should_record_variables:
                    self.push_traceable_variables(frame, current.locals)
                self.add_to_event_buffer(current)

        if event == EventKeys.call:
            cursor = self.create_cursor(file_path_under_cursor, frame)
            self.call_stack.enter_frame(cursor)
            # lets try to add methods
            # [if sampling mode]
            if will_record_current_event:
                current = events.MethodEnterEvent(cursor, self.get_execution_stack(), now)
                if self.should_record_variables:
                    variables = current.input_variables
                    self.push_traceable_variables(frame, variables)
                self.add_to_event_buffer(current)

        if event == EventKeys.event_return:
            self.call_stack.exit_frame()
            # [? is sampling]
            if will_record_current_event:
                cursor = self.create_cursor(file_path_under_cursor, frame)
                current = events.MethodExitEvent(cursor, self.get_execution_stack(), now)
                if self.should_record_variables:
                    current.return_variables.push_variable('__return', arg)
                    self.push_traceable_variables(frame, current.locals)
                self.add_to_event_buffer(current)

        self.flush_queue_if_full()

    def add_to_event_buffer(self, current):
        # todo: is this caused because of array dynamic size/doubling?
        # if True:
        #     return
        self.event_buffer.add_event(current)

    def create_cursor(self, file_path_under_cursor, frame):
        file_id = self.file_map.file_id(file_path_under_cursor)
        cursor = events.ExecutionCursor(file_id, frame.f_lineno, frame.f_code.co_name)
        return cursor

    def get_execution_stack(self):
        return self.call_stack.current_frame()

    def push_traceable_variables(self, frame, variables):
        for (name, value) in frame.f_locals.items():
            # todo use variable values diffing
            variables.push_variable(name, value)

    def flush_outstanding_events(self):
        old_buffer = self.event_buffer.finish_chunk()
        self.perf.print_avg_time()

        self.queue.put_events(EventsSlice(self.session.session_name, self.event_number, old_buffer, self.file_map.files.copy()))
        self.event_number += 1

    def flush_queue_if_full(self):
        if self.event_buffer.how_many_events() >= self.max_events_before_send:
            self.flush_outstanding_events()



import threading
from queue import Queue, Empty

from pycrunch_trace.client.networking.commands import EventsSlice, StopCommand, AbstractNetworkCommand, StartCommand, FileContentSlice

import sys
import pyximport

from pycrunch_trace.client.networking.strategies.abstract_strategy import AbstractRecordingStrategy
from pycrunch_trace.file_system.trace_session import TraceSession

pyximport.install()
from pycrunch_trace.client.networking.strategies.native_write_strategy import NativeLocalRecordingStrategy


import logging


logger = logging.getLogger(__name__)

# root = logging.getLogger()
# root.setLevel(logging.DEBUG)
#
# handler = logging.StreamHandler(sys.stdout)
# handler.setLevel(logging.DEBUG)
# formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
# handler.setFormatter(formatter)
# root.addHandler(handler)
# root.getChild('engineio.client').disabled = True
# root.getChild('socketio.client').disabled = True
# logger = logging.getLogger(__name__)


import os

class ClientQueueThread:
    available_recording_strategies = ['network', 'local']

    is_thread_running: bool
    _counter: int
    _strategy: AbstractRecordingStrategy

    def __init__(self):
        print(f'PID: {os.getpid()} ClientQueueThread init')
        self._counter = 0
        self.so_far = 0
        self.is_connected = False
        self.outgoingQueue = Queue()
        self.is_thread_running = False
        current_strategy = 'native_local'
        if current_strategy == 'native_local':
            self._strategy = NativeLocalRecordingStrategy()

    def tracing_will_start(self, session_id: str):
        self.ensure_thread_started()
        try:
            self.outgoingQueue.put_nowait(StartCommand(session_id))
        except Exception as e:
            print('EXCEPTION')
            print(e)


    def put_events(self, events: EventsSlice):
        self.so_far +=  len(events.events)
        print(f'{events.session_id} - put_events: so far: {self.so_far}')
        self.ensure_thread_started()
        try:
            self.outgoingQueue.put_nowait(events)
        except Exception as e:
            print('EXCEPTION while put_events')
            print(e)

    def put_file_slice(self, events: FileContentSlice):
        print('put_file_slice')
        self.ensure_thread_started()
        try:
            self.outgoingQueue.put_nowait(events)
        except Exception as e:
            print('EXCEPTION while put_file_slice')
            print(e)

    def tracing_did_complete(self, session_id, session: TraceSession):
        print('tracing_did_complete')
        self.ensure_thread_started()
        self.outgoingQueue.put_nowait(
            StopCommand(
                session_id,
                list(session.files_in_session.copy()),
                list(session.excluded_files.copy()),
            )
        )

    def start(self):
        print(f'start thread dispather queue with pid: {os.getpid()}')
        if self.is_thread_running:
            return

        print('socketio init')
        x = threading.Thread(target=self.thread_proc, args=(42,))
        # x.setDaemon(True)
        x.setDaemon(False)
        x.start()
        # todo lock?
        self.is_thread_running = True


    def thread_proc(self, obj):
        logging.info("Thread ClientQueueThread::thread_proc: starting")

        self._strategy.prepare()


        while True:
            logger.info('outgoingQueue.get: Waiting for message...')
            try:
                x: AbstractNetworkCommand = self.outgoingQueue.get(True, 3)
                print(f'queue length {len(self.outgoingQueue.queue)}')

                if x is not None:
                    self.process_single_message(x)
            except Empty:
                print('Timeout while waiting for new msg... Thread will stop for now')
                break
                pass

            except Exception as ex:
                logger.info('Ex while getting message from queue')
                print('===!!! Ex while getting message from queue')
                print(str(ex))
                exc_type, exc_obj, exc_tb = sys.exc_info()
                fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
                print(exc_type, fname, exc_tb.tb_lineno)
                continue
        # end while
        print('Thread stopped')
        self._strategy.clean()
        self.is_thread_running = False

    def process_single_message(self, x: AbstractNetworkCommand):
        print(f'got evt {x.command_name}')
        if x.command_name == 'StartCommand':
            self._strategy.recording_start(x.session_id)
        if x.command_name == 'StopCommand':
            logger.info('got ' + x.command_name)
            self._strategy.recording_stop(x.session_id, x.files_included, x.files_excluded)
        if x.command_name == 'FileContentSlice':
            logger.info('got ' + x.command_name)
            self._strategy.files_slice(x)
        logger.info('Sending... ' + x.command_name)
        if x.command_name == 'EventsSlice':
            self._strategy.recording_slice(x)
            logger.info('Sent... ' + x.command_name)

    def ensure_thread_started(self):
        if not self.is_thread_running:
            self.start()


event_queue: ClientQueueThread = ClientQueueThread()



class TracingServerEvent:
    event_name: str


class RecordingStartEvent(TracingServerEvent):
    session_id: str

    def __init__(self, session_id: str):
        self.session_id = session_id
        self.event_name = 'recording_start'


class RecordingCompleteEvent(TracingServerEvent):
    session_id: str

    def __init__(self, session_id: str):
        self.session_id = session_id
        self.event_name = 'recording_complete'


class PartialFileChunkEvent(TracingServerEvent):
    def __init__(self, event_number, session_id, bytes_to_write, events_in_payload):
        self.events_in_payload = events_in_payload
        self.bytes_to_write = bytes_to_write
        self.session_id = session_id
        self.event_number = event_number
        self.event_name = 'recording_chunk'



import collections
from collections import defaultdict
from typing import List, Dict

from pycrunch_trace.tracing.file_map import FileMap


class ClientTraceIntrospection:
    total_events: int

    def __init__(self):
        self.total_events = 0
        self.stats = defaultdict(int)
        # file id -> hit count
        self.top_files = defaultdict(int)

    def save_events(self, events: List):
        self.total_events += len(events)
        for e in events:
            self.stats[e.event_name] += 1
            self.top_files[e.cursor.file] += 1

    def print_to_console(self, files: Dict[str, int]):
        print('TraceIntrospection:')
        print('  stats:')
        for (each, hit_count) in self.stats.items():
            print(f' - {each}:{hit_count}')

        print('  files:')
        filemap = FileMap.from_reverse(files)
        sorted_x = sorted(self.top_files.items(), reverse=True, key=lambda kv: kv[1])
        sortir = collections.OrderedDict(sorted_x)


        for (each, hit_count) in sortir.items():
            print(f' - {hit_count} hits in {filemap.filename_by_id(each)}')


client_introspection = ClientTraceIntrospection()



from pathlib import Path
from typing import Union

from .file import File


class Directory:
    path: Path

    def __init__(self, path: Union[str, Path]):
        if type(path) == str:
            self.path = Path(path)
        elif isinstance(path, Path):
            self.path = path
        else:
            raise Exception('Not supported path: ' + str(path))

    def folders(self):
        result = []
        self._ensure_created()
        for maybe_be_folder in self.path.glob('*'):
            if maybe_be_folder.is_dir():
                result.append(maybe_be_folder.name)
        return result

    def files(self, extension: str):
        if not self._exists():
            return []

        result = []
        for file in self.path.glob(f'*.{extension}'):
            result.append(File(file))
        return result

    def _ensure_created(self):
        if not self._exists():
            self.path.mkdir(exist_ok=True)

    def _exists(self):
        return self.path.exists()

    def __str__(self):
        return str(self.path)



import yaml

from . import AbstractFileFilter
from ..oop import AbstractFile


class CustomFileFilter(AbstractFileFilter):
    exclusions: tuple
    _trace_variables: bool
    profile_file: AbstractFile
    _loaded: bool

    def __init__(self, profile_file: AbstractFile):
        self.profile_file = profile_file
        self._loaded = False
        self._trace_variables = True
        self.exclusions = ()

    def all_exclusions(self):
        self._ensure_loaded()
        return list(self.exclusions)

    def should_trace(self, filename: str):
        self._ensure_loaded()
        if filename.startswith(self.exclusions) or filename.endswith(self.exclusions):
            # print('should_trace: false - filename= '+filename)
            return False
        return True

    def _ensure_loaded(self):
        if self._loaded:
            return

        self._load()
        self._loaded = True

    def _load(self):
        all = yaml.load(self.profile_file.as_bytes(), Loader=yaml.FullLoader)
        self.load_exclusions(all)
        self.load_variables(all)

    def load_exclusions(self, all):
        tmp = set()
        exclusions = all.get('exclusions')
        if exclusions:
            for e in exclusions:
                tmp.add(e)
        self.exclusions = tuple(tmp)

    def load_variables(self, all):
        trace_vars = all.get('trace_variables')
        if isinstance(trace_vars, bool):
            self._trace_variables = trace_vars
            if not self._trace_variables:
                print('!! PyCrunch - Variables will not be recorded in session')

    def should_record_variables(self) -> bool:
        self._ensure_loaded()
        return self._trace_variables




from pathlib import Path
from typing import List

from pycrunch_trace.config import config
from pycrunch_trace.file_system.persisted_session import PersistedSession, LazyLoadedSession


class SessionStore:
    recording_directory: Path

    def __init__(self):
        self.recording_directory = config.recording_directory
        pass

    def all_sessions(self) -> List[str]:
        result = []
        self.ensure_recording_directory_created()
        for maybe_be_folder in self.recording_directory.glob('*'):
            if maybe_be_folder.is_dir():
                result.append(maybe_be_folder.name)
        return result

    def load_session(self, session_name: str) -> LazyLoadedSession:
        self.ensure_recording_directory_created()
        load_from = self.recording_directory.joinpath(session_name)
        session = PersistedSession.load_from_directory(load_from)
        return session

    def new_session(self, session_name) -> PersistedSession:
        self.ensure_recording_directory_created()
        session_directory = self.create_session_directory(session_name)
        return PersistedSession(session_directory)

    def create_session_directory(self, session_name):
        session_directory = self.recording_directory.joinpath(session_name)
        self.ensure_directory_created(session_directory)
        return session_directory

    def ensure_recording_directory_created(self):
        self.ensure_directory_created(self.recording_directory)

    def ensure_directory_created(self, directory: Path):
        if not directory.exists():
            directory.mkdir(exist_ok=True)




# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: message.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='message.proto',
  package='',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=b'\n\rmessage.proto\"d\n\x0cTraceSession\x12\x1b\n\x06\x65vents\x18\x01 \x03(\x0b\x32\x0b.TraceEvent\x12!\n\x0cstack_frames\x18\x02 \x03(\x0b\x32\x0b.StackFrame\x12\x14\n\x05\x66iles\x18\x03 \x03(\x0b\x32\x05.File\"-\n\x0e\x46ilesInSession\x12\x1b\n\x05\x66iles\x18\x01 \x03(\x0b\x32\x0c.FileContent\"*\n\x0b\x46ileContent\x12\n\n\x02id\x18\x01 \x01(\x05\x12\x0f\n\x07\x63ontent\x18\x02 \x01(\x0c\"\xc7\x01\n\nTraceEvent\x12\x12\n\nevent_name\x18\x01 \x01(\t\x12 \n\x06\x63ursor\x18\x02 \x01(\x0b\x32\x10.ExecutionCursor\x12\x10\n\x08stack_id\x18\x03 \x01(\x05\x12#\n\x0finput_variables\x18\x04 \x01(\x0b\x32\n.Variables\x12\x1a\n\x06locals\x18\x05 \x01(\x0b\x32\n.Variables\x12$\n\x10return_variables\x18\x06 \x01(\x0b\x32\n.Variables\x12\n\n\x02ts\x18\x07 \x01(\x02\")\n\tVariables\x12\x1c\n\tvariables\x18\x01 \x03(\x0b\x32\t.Variable\"\'\n\x08Variable\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\"D\n\x0f\x45xecutionCursor\x12\x0c\n\x04\x66ile\x18\x01 \x01(\x05\x12\x0c\n\x04line\x18\x02 \x01(\x05\x12\x15\n\rfunction_name\x18\x03 \x01(\t\" \n\x04\x46ile\x12\n\n\x02id\x18\x01 \x01(\x05\x12\x0c\n\x04\x66ile\x18\x02 \x01(\t\"^\n\nStackFrame\x12\n\n\x02id\x18\x01 \x01(\x05\x12\x0c\n\x04\x66ile\x18\x02 \x01(\x05\x12\x0c\n\x04line\x18\x03 \x01(\x05\x12\x11\n\tparent_id\x18\x04 \x01(\x05\x12\x15\n\rfunction_name\x18\x05 \x01(\tb\x06proto3'
)




_TRACESESSION = _descriptor.Descriptor(
  name='TraceSession',
  full_name='TraceSession',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='events', full_name='TraceSession.events', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='stack_frames', full_name='TraceSession.stack_frames', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='files', full_name='TraceSession.files', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=17,
  serialized_end=117,
)


_FILESINSESSION = _descriptor.Descriptor(
  name='FilesInSession',
  full_name='FilesInSession',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='files', full_name='FilesInSession.files', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=119,
  serialized_end=164,
)


_FILECONTENT = _descriptor.Descriptor(
  name='FileContent',
  full_name='FileContent',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='FileContent.id', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='content', full_name='FileContent.content', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=166,
  serialized_end=208,
)


_TRACEEVENT = _descriptor.Descriptor(
  name='TraceEvent',
  full_name='TraceEvent',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='event_name', full_name='TraceEvent.event_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='cursor', full_name='TraceEvent.cursor', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='stack_id', full_name='TraceEvent.stack_id', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='input_variables', full_name='TraceEvent.input_variables', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='locals', full_name='TraceEvent.locals', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='return_variables', full_name='TraceEvent.return_variables', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ts', full_name='TraceEvent.ts', index=6,
      number=7, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=211,
  serialized_end=410,
)


_VARIABLES = _descriptor.Descriptor(
  name='Variables',
  full_name='Variables',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='variables', full_name='Variables.variables', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=412,
  serialized_end=453,
)


_VARIABLE = _descriptor.Descriptor(
  name='Variable',
  full_name='Variable',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='Variable.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='Variable.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=455,
  serialized_end=494,
)


_EXECUTIONCURSOR = _descriptor.Descriptor(
  name='ExecutionCursor',
  full_name='ExecutionCursor',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='file', full_name='ExecutionCursor.file', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='line', full_name='ExecutionCursor.line', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='function_name', full_name='ExecutionCursor.function_name', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=496,
  serialized_end=564,
)


_FILE = _descriptor.Descriptor(
  name='File',
  full_name='File',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='File.id', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='file', full_name='File.file', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=566,
  serialized_end=598,
)


_STACKFRAME = _descriptor.Descriptor(
  name='StackFrame',
  full_name='StackFrame',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='StackFrame.id', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='file', full_name='StackFrame.file', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='line', full_name='StackFrame.line', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='parent_id', full_name='StackFrame.parent_id', index=3,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='function_name', full_name='StackFrame.function_name', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=600,
  serialized_end=694,
)

_TRACESESSION.fields_by_name['events'].message_type = _TRACEEVENT
_TRACESESSION.fields_by_name['stack_frames'].message_type = _STACKFRAME
_TRACESESSION.fields_by_name['files'].message_type = _FILE
_FILESINSESSION.fields_by_name['files'].message_type = _FILECONTENT
_TRACEEVENT.fields_by_name['cursor'].message_type = _EXECUTIONCURSOR
_TRACEEVENT.fields_by_name['input_variables'].message_type = _VARIABLES
_TRACEEVENT.fields_by_name['locals'].message_type = _VARIABLES
_TRACEEVENT.fields_by_name['return_variables'].message_type = _VARIABLES
_VARIABLES.fields_by_name['variables'].message_type = _VARIABLE
DESCRIPTOR.message_types_by_name['TraceSession'] = _TRACESESSION
DESCRIPTOR.message_types_by_name['FilesInSession'] = _FILESINSESSION
DESCRIPTOR.message_types_by_name['FileContent'] = _FILECONTENT
DESCRIPTOR.message_types_by_name['TraceEvent'] = _TRACEEVENT
DESCRIPTOR.message_types_by_name['Variables'] = _VARIABLES
DESCRIPTOR.message_types_by_name['Variable'] = _VARIABLE
DESCRIPTOR.message_types_by_name['ExecutionCursor'] = _EXECUTIONCURSOR
DESCRIPTOR.message_types_by_name['File'] = _FILE
DESCRIPTOR.message_types_by_name['StackFrame'] = _STACKFRAME
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

TraceSession = _reflection.GeneratedProtocolMessageType('TraceSession', (_message.Message,), {
  'DESCRIPTOR' : _TRACESESSION,
  '__module__' : 'message_pb2'
  # @@protoc_insertion_point(class_scope:TraceSession)
  })
_sym_db.RegisterMessage(TraceSession)

FilesInSession = _reflection.GeneratedProtocolMessageType('FilesInSession', (_message.Message,), {
  'DESCRIPTOR' : _FILESINSESSION,
  '__module__' : 'message_pb2'
  # @@protoc_insertion_point(class_scope:FilesInSession)
  })
_sym_db.RegisterMessage(FilesInSession)

FileContent = _reflection.GeneratedProtocolMessageType('FileContent', (_message.Message,), {
  'DESCRIPTOR' : _FILECONTENT,
  '__module__' : 'message_pb2'
  # @@protoc_insertion_point(class_scope:FileContent)
  })
_sym_db.RegisterMessage(FileContent)

TraceEvent = _reflection.GeneratedProtocolMessageType('TraceEvent', (_message.Message,), {
  'DESCRIPTOR' : _TRACEEVENT,
  '__module__' : 'message_pb2'
  # @@protoc_insertion_point(class_scope:TraceEvent)
  })
_sym_db.RegisterMessage(TraceEvent)

Variables = _reflection.GeneratedProtocolMessageType('Variables', (_message.Message,), {
  'DESCRIPTOR' : _VARIABLES,
  '__module__' : 'message_pb2'
  # @@protoc_insertion_point(class_scope:Variables)
  })
_sym_db.RegisterMessage(Variables)

Variable = _reflection.GeneratedProtocolMessageType('Variable', (_message.Message,), {
  'DESCRIPTOR' : _VARIABLE,
  '__module__' : 'message_pb2'
  # @@protoc_insertion_point(class_scope:Variable)
  })
_sym_db.RegisterMessage(Variable)

ExecutionCursor = _reflection.GeneratedProtocolMessageType('ExecutionCursor', (_message.Message,), {
  'DESCRIPTOR' : _EXECUTIONCURSOR,
  '__module__' : 'message_pb2'
  # @@protoc_insertion_point(class_scope:ExecutionCursor)
  })
_sym_db.RegisterMessage(ExecutionCursor)

File = _reflection.GeneratedProtocolMessageType('File', (_message.Message,), {
  'DESCRIPTOR' : _FILE,
  '__module__' : 'message_pb2'
  # @@protoc_insertion_point(class_scope:File)
  })
_sym_db.RegisterMessage(File)

StackFrame = _reflection.GeneratedProtocolMessageType('StackFrame', (_message.Message,), {
  'DESCRIPTOR' : _STACKFRAME,
  '__module__' : 'message_pb2'
  # @@protoc_insertion_point(class_scope:StackFrame)
  })
_sym_db.RegisterMessage(StackFrame)


# @@protoc_insertion_point(module_scope)



from time import sleep

from pycrunch_trace.tracing.inline_profiler import ProfilingScope, inline_profiler_instance


def call_some_funtion(i):
    with ProfilingScope('call_some_funtion'):
        sleep(0.1)
        pass


def test_nesting():
    with ProfilingScope('root'):
        for i in range(10):
            with ProfilingScope('loop'):
                call_some_funtion(i)
    inline_profiler_instance.print_timings()


import asyncio
import io

# import aiohttp_debugtoolbar
import logging.config

from pathlib import Path

import yaml
from aiohttp import web
import socketio
from pycrunch.api.shared import sio
from pycrunch_trace.config import config

import cgitb
cgitb.enable(format='text')





package_directory = Path(__file__).parent
print(package_directory)
engine_directory = package_directory.parent
config.set_package_directory(package_directory)
config.set_engine_directory(engine_directory)
configuration_yaml_ = package_directory.joinpath('logging-configuration.yaml')
print(configuration_yaml_)
with open(configuration_yaml_, 'r') as f:
    logging.config.dictConfig(yaml.safe_load(f.read()))


import logging

logger = logging.getLogger(__name__)



import sys

if sys.platform == 'win32':
    policy = asyncio.get_event_loop_policy()
    policy._loop_factory = asyncio.ProactorEventLoop

async def shutdown(loop, signal=None):
    print('PTrace Server: shutdown')
    """Cleanup tasks tied to the service's shutdown."""
    if signal:
        logging.info(f"Received exit signal {signal.name}...")
    logging.info("Closing database connections")

def handle_exception(loop, context):
    # context["message"] will always be there; but context["exception"] may not
    print('PTrace Server: !!!PISOS')
    print('PTrace Server: handle_exception')
    msg = context.get("exception", context["message"])
    print('error was : ' + str(msg))

    logging.error(f"Caught exception: {msg}")
    logging.info("Shutting down...")
    # asyncio.create_task(shutdown(loop))

def run():
    loop = asyncio.get_event_loop()
    loop.set_debug(True)
    loop.set_exception_handler(handle_exception)

    app = web.Application()
    # aiohttp_debugtoolbar.setup(app)


    #  keep import outside main
    from pycrunch_trace.server.shared import tracer_socket_server

    # attach sio events
    # noinspection PyUnresolvedReferences
    import pycrunch_trace.server.recording_server_websocket

    tracer_socket_server.attach(app)


    async def index(request):
        """Serve the client-side application."""
        with io.open(config.absolute_path, 'r') as f:
            lines = f.read()
            return web.Response(text=lines, content_type='application/json')


    # app.router.add_static('/static', 'static')
    app.router.add_get('/', index)
    web.run_app(app)

if __name__ == '__main__':
   run()


import requests

from pycrunch_trace.client.api import trace


def some_code():
    for x in range(1):
        req = requests.get('https://google.com')
        code = req.status_code
        print(str(x))
        print(code)

@trace
def run_youtube_code():
    some_code()


run_youtube_code()


import io
import os
import struct
from pathlib import Path
from tempfile import NamedTemporaryFile

from pycrunch_trace.client.api import trace
from pycrunch_trace.file_system.trace_file import TraceFile


def test_trace_file_header():
    f = NamedTemporaryFile(prefix='pytracer', delete=False)
    name = f.name
    f.close()
    print(name)
    try:
        sut = TraceFile('dummy', Path(f.name))

        x = sut.write_header_placeholder()
        sut.update_file_header_files_section(123)
        with io.FileIO(name, 'r') as actual:
            # SIGNATURE
            buffer = actual.read(struct.calcsize('>i'))
            sig = struct.unpack('>i', buffer)[0]
            assert sig == 15051991

            buffer = actual.read(struct.calcsize('>i'))
            tag = struct.unpack('>i', buffer)[0]
            assert tag == 1

            buffer = actual.read(struct.calcsize('>i'))
            header_size = struct.unpack('>i', buffer)[0]
            assert header_size == 16 * 1024


            new_pos = actual.seek(header_size + 8 + 4, io.SEEK_SET)
            print(new_pos)
            # SIGNATURE
            buffer = actual.read(struct.calcsize('>i'))
            sig = struct.unpack('>i', buffer)[0]
            assert sig == 15051991

    finally:
        os.remove(f.name)
        pass
    pass


from pycrunch_trace.client.api.trace import Trace


def xxx(x: int):
    y = x * 9
    xy = x * 2
    zalupa = dict(x=1, y='some_data')
    for i in range(xy):
        xy += i
        zalupa[str(i)] = i * 8
    return xy


yoba = Trace()
yoba.start(session_name='network_demo3')


def run(param):
    param(4)


run(xxx)


yoba.stop()

print('stoped')



# from pycrunch_trace.events.size_prediction import SizeBreakdown


# @pytest.mark.skip()
# def test_1():
#     x = SizeBreakdown.load_from_session()
import random
import string

from pycrunch.insights import trace

from pycrunch_trace.file_system.human_readable_size import HumanReadableByteSize


def randomString(stringLength=10):
    """Generate a random string of fixed length """
    letters = string.ascii_lowercase
    return ''.join(random.choice(letters) for i in range(stringLength))


class Repeat(object):
    def __init__(self, times):
        self.times = times

    def  __str__(self):
        res =  ''
        for i in range(self.times):
            res += '/' + randomString(20)
        return res


def random_filename():
    return 'users/' + randomString(5) + '/projects/' + randomString(20) + '/' + randomString(20) + Repeat(random.randint(1, 6)).__str__() + '.py'


class NewEvent:
    event_names = ['method_enter', 'line', 'method_exit']

    def __init__(self, filename):
        self.name = random.choice(OldEvent.event_names)
        self.line = random.randint(1, 1000)
        self.filename = filename
        self.function_name = randomString(random.randint(1,20))

    def size(self):
        function_name_size = len(self.function_name)

        call_stack_length = 8 + 8 + 8 + function_name_size + 8
        method_name_size = len(self.name)
        method_name_size = 0
        return call_stack_length + method_name_size + 8 + 8 + function_name_size


class OldEvent:
    event_names = ['method_enter', 'line', 'method_exit']

    def __init__(self, filename):
        self.name = random.choice(OldEvent.event_names)
        self.line = random.randint(1, 1000)
        self.filename = filename
        self.function_name = randomString(random.randint(1,20))

    def size(self):
        call_stack_length = 8 + 8 + len(self.name) + len(self.filename)
        return call_stack_length + len(self.name) + 8 + len(self.filename) + len(self.function_name)

def test_old_size():
    total_files = 255
    files_in_session = []
    for x in range(total_files):
        files_in_session.append(random_filename())
    # trace(files_in_session=files_in_session)
    total = 0
    for x in range(10000):
        total += size_of_random_event(files_in_session)
    # 10000 ='3.2 MB'
    # 100000 = 31 MB
    # 1 000 000 = 310 MB
    str__ = HumanReadableByteSize(total).__str__()
    print(str__)
    trace(total=str__)

def test_new_size():
    total_files = 255
    files_in_session = []
    size_files = 0
    for x in range(total_files):

        filename = random_filename()
        size_files += len(files_in_session)
        files_in_session.append(filename)

    trace(size_files=size_files)

    # trace(files_in_session=files_in_session)
    total = 0
    for x in range(10000):
        total += size_of_new_random_event(files_in_session)
    # 10000 ='3.2 MB'
    # 100000 = 31 MB
    # 1 000 000 = 310 MB
    str__ = HumanReadableByteSize(total).__str__()
    print(str__)
    trace(total=str__)

def size_of_random_event(files_in_session):
    return OldEvent(random.choice(files_in_session)).size()

def size_of_new_random_event(files_in_session):
    return NewEvent(random.choice(files_in_session)).size()

def test_232():
    print('s')


from pycrunch.insights import trace

import pycrunch_trace.tracing.simulation.models as sim
from pycrunch_trace.serialization import to_string
from pycrunch_trace.session.snapshot import snapshot
from pycrunch_trace.tracing.simple_tracer import SimpleTracer

absolute_path = '/pycrunch_tracer/samples/module_a.py'

def test_sim1():
    def code_method() -> sim.Code:
        code = sim.Code()
        code.co_filename = absolute_path
        code.co_name = 'some_method'
        code.co_argcount = 1
        code.co_argcount = 1
        return code

    method = code_method()

    def get_frame(line: int):
        frame = sim.Frame()
        frame.f_lineno = line
        frame.f_code = method
        frame.f_locals = dict(some_number=1)
        return frame
    event_buffer = []
    sut = SimpleTracer(event_buffer, 'testing',,
    sut.simple_tracer(get_frame(2), sim.EventKeys.call, None)
    sut.simple_tracer(get_frame(3), sim.EventKeys.line, None)
    sut.simple_tracer(get_frame(4), sim.EventKeys.line, None)

    sut.simple_tracer(get_frame(2), sim.EventKeys.call, None)
    sut.simple_tracer(get_frame(3), sim.EventKeys.line, None)
    sut.simple_tracer(get_frame(2), sim.EventKeys.call, None)
    sut.simple_tracer(get_frame(3), sim.EventKeys.line, None)
    sut.simple_tracer(get_frame(4), sim.EventKeys.line, None)
    sut.simple_tracer(get_frame(4), sim.EventKeys.event_return, None)
    sut.simple_tracer(get_frame(4), sim.EventKeys.line, None)
    sut.simple_tracer(get_frame(4), sim.EventKeys.event_return, None)

    sut.simple_tracer(get_frame(5), sim.EventKeys.line, None)
    sut.simple_tracer(get_frame(6), sim.EventKeys.line, None)
    sut.simple_tracer(get_frame(6), sim.EventKeys.event_return, 6)

    coe = sut.simulation.simulated_code()
    trace(coe=coe)
    trace(sixt=to_string(event_buffer[6]))
    trace(buffer=to_string(event_buffer))
    snapshot.save('a', event_buffer)
    # trace(x=x)
    pass









import functools
import sys

from ppretty import ppretty


def trace_lines(frame, event, arg):
    if event == 'return':
        print(f' -> return')
        print('f_locals ' + ppretty(frame))

        return

    if event != 'line':
        return
    co = frame.f_code
    func_name = co.co_name
    line_no = frame.f_lineno
    print(f'*  {func_name} line {line_no}')
    print(f'*    locals: {frame.f_locals}')


def trace_calls(frame, event, arg, to_be_traced):
    if event != 'call':
        return
    co = frame.f_code
    func_name = co.co_name
    if func_name == 'write':
        # Ignore write() calls from printing
        return
    line_no = frame.f_lineno
    filename = co.co_filename
    if not filename.endswith('sys_settrace_line.py'):
        # Ignore calls not in this module
        return
    print('* Call to {} on line {} of {}'.format(
        func_name, line_no, filename))
    if func_name in to_be_traced:
        # Trace into this function
        return trace_lines
    return


def c(input):
    print('input =', input)
    print('Leaving c()')


def b(arg):
    val = arg * 5
    val = val * 2
    c(val)
    print('Leaving b()')
    return 42


def a():
    b(2)
    print('Leaving a()')


tracer = functools.partial(trace_calls, to_be_traced=['b'])
sys.settrace(tracer)
a()


import io
from pathlib import Path
from typing import Union

class AbstractFile:
    def as_bytes(self):
        pass

    def short_name(self):
        pass

class File(AbstractFile):
    class Mock(AbstractFile):
        def __init__(self, buffer: bytes, filename: str = None):
            if filename:
                self.filename = filename
            else:
                self.filename = 'mock'

            self.buffer = buffer

        def as_bytes(self):
            return self.buffer

        def short_name(self):
            return self.buffer
    filename: str

    def __init__(self, filename: Union[str, Path]):
        if type(filename) == str:
            self.filename = filename
        elif isinstance(filename, Path):
            self.filename = str(filename)

    def as_bytes(self):
        with io.FileIO(self.filename, mode='r') as current_file:
            return current_file.readall()

    def short_name(self) -> str:
        return Path(self.filename).name

    def __str__(self):
        return self.filename




from pycrunch_trace.client.api import trace, Trace
from pycrunch_trace.oop import Clock


def nested_function(input_a2):
    input_a2 += 2
    return input_a2

def another_function(input_argument):
    x = nested_function(input_argument)
    y = nested_function(input_argument + 1)
    return x + y

def million_of_calls():
    # 20000 = 300008
    max_events = 10
    max_events = 100000
    for x in range(max_events):
        another_function(x)

clock = Clock()
my_yoba = Trace()
my_yoba.start('million_of_calls')
start = clock.now()

print('before', start)
million_of_calls()
end = clock.now()
print('after', end)
print('diff', end - start)

my_yoba.stop()
# print(my_yoba._tracer.simulation.simulated_code())



from pathlib import Path


class TracerConfig:
    absolute_path = '/pycrunch_tracer/samples/module_b.py'
    def __init__(self):
        self.engine_directory = None
        self.recording_directory = self.get_default_recording_directory()

        self.package_directory = None

    def get_default_recording_directory(self):
        # return Path('/Volumes/WD_BLACK/pycrunch-trace')
        # return Path('/Volumes/Kingston/pycrunch-trace')
        return Path.joinpath(Path(__file__).parent, 'pycrunch-recordings')

    def set_engine_directory(self, engine_directory: str):
        self.engine_directory = engine_directory

    def set_package_directory(self, package_directory: str):
        self.package_directory = package_directory

config = TracerConfig()




from typing import List, Optional


class LiveSession:
    sid: str
    session_name: str
    version: str

    def __init__(self, sid, version):
        self.sid = sid
        self.version = version


class ActiveConnections:
    tracers_online: List[LiveSession]
    clients: List[str]

    def __init__(self):
        self.tracers_online = []
        self.clients = []

    def client_did_connect(self, reference: str):
        self.clients.append(reference)

    def client_did_disconnect(self, reference: str):
        self.clients.remove(reference)
        self.remove_if_any_tracer(reference)

    def remove_if_any_tracer(self, reference: str) -> bool:
        possible_tracer = self.find_tracer_with_id(reference)
        if not possible_tracer:
            return False

        self.tracers_online.remove(possible_tracer)
        return True

    def find_tracer_with_id(self, reference) -> Optional[LiveSession]:
        for tracer in self.tracers_online:
            if tracer.sid == reference:
                return tracer

        return None

    def tracer_did_connect(self, sid, version):
        self.tracers_online.append(LiveSession(sid, version))
        pass

    def tracer_did_disconnect(self, sid):
        return self.remove_if_any_tracer(sid)




import io
import os
import struct
from pathlib import Path

from pycrunch_trace.file_system.persisted_session import PersistedSession
from pycrunch_trace.file_system.session_store import SessionStore


def test_write():
    x = SessionStore()
    x.ensure_recording_directory_created()
    directory__joinpath = Path(x.recording_directory).joinpath('session_fake')
    x.ensure_directory_created(directory__joinpath)

    target_file = directory__joinpath.joinpath(PersistedSession.chunked_recording_filename)
    os.remove(target_file)

    write_once(target_file, 1)
    write_once(target_file, 2)
    write_once(target_file, 3)

def test_read():
    x = SessionStore()
    x.ensure_recording_directory_created()
    directory__joinpath = Path(x.recording_directory).joinpath('session_fake')
    x.ensure_directory_created(directory__joinpath)

    header_size = struct.calcsize("i")
    target_file = directory__joinpath.joinpath(PersistedSession.chunked_recording_filename)
    with io.FileIO(target_file, 'r') as file_to_read:
        while True:
            header_bytes = file_to_read.read(header_size)
            if len(header_bytes) <= 0:
                print(f' -- Read to end')

                break
            next_chunk_length = struct.unpack('i', header_bytes)[0]
            print(f'next_chunk_length {next_chunk_length}')
            read_bytes = file_to_read.read(next_chunk_length)
            result = read_bytes.decode('utf-8')
            print(f'result {result}')

    #
    # write_once(target_file, 1)
    # write_once(target_file, 2)
    # write_once(target_file, 3)


def write_once(target_file, index):
    target_mode = 'a'
    if not target_file.exists():
        target_mode = 'w'
    with io.FileIO(target_file, target_mode) as file_to_write:
        bytes_to_write = b'zalupa' + str(index).encode('utf-8')

        length_of_message = len(bytes_to_write)
        header_bytes = struct.pack("i", length_of_message)

        file_to_write.write(header_bytes + bytes_to_write)



import sys
from time import sleep

from pycrunch_trace.client.api import trace


@trace('timings_precision')
def run(x):
    y = x
    sleep(0.3)
    b()
    info = str(sys.float_info)
    print(info)
    sleep(1)


def x():
    sleep(0.1)


def b():
    y = 'sleeped 01'
    sleep(0.2)
    for i in range(10):
        x()
    c()


def c():
    y = 'sleeped 01'
    sleep(0.7)
    y = 'sleeped 1 sec total'


run(1)




from pycrunch_trace import config
from pycrunch_trace.filters import can_trace_type


def test_can_trace_int():
    assert can_trace_type(1)


def test_can_trace_float():
    assert can_trace_type(1.5)

def test_can_trace_bool():
    assert can_trace_type(True)

def test_can_trace_none():
    assert can_trace_type(None)

def test_can_trace_string():
    assert can_trace_type('some_str')

def test_can_trace_dict():
    assert can_trace_type(dict(kwargs=dict(a=1)))

def test_no_trace_module():
    assert not can_trace_type(config)



from typing import TypeVar

from .base_event import Event
from ..filters import can_trace_type


class ExecutionCursor:
    function_name: str
    file: int
    line: int

    def __init__(self, file: int, line: int, function_name: str):
        self.function_name = function_name
        self.file = file
        self.line = line

class StackFrameIds:
    last_id: int
    def __init__(self):
        self.last_id = 1

    def new_id(self):
        ret = self.last_id
        self.last_id += 1
        return ret

stack_ids = StackFrameIds()

class StackFrame:
    # parent: StackFrame
    def __init__(self, parent, file: id, line: int, function_name: str):
        self.parent = parent
        self.file = file
        self.line = line
        self.id = stack_ids.new_id()
        self.function_name = function_name

    def as_id(self):
        return self.id


    @classmethod
    def new(cls, parent, execution_cursor: ExecutionCursor):
        return StackFrame(parent, execution_cursor.file, execution_cursor.line, execution_cursor.function_name)

    @classmethod
    def clone(cls, origin):
        if not origin:
            return StackFrame.empty()
        return StackFrame(origin.parent, origin.file, origin.line, origin.function_name)

    @classmethod
    def empty(cls):
        return StackFrame(None, None, None, None)

    def __str__(self):
        return f'{self.file}:{self.function_name}:{self.line} -> \n\t {self.parent}'




class Variables:
    variables: dict

    def __init__(self):
        # self.variables = None
        # return
        self.variables = dict()

    def push_variable(self, name, value):
        self.variables[name] = self.ensure_safe_for_serialization(value)

    def ensure_safe_for_serialization(self, value):
        # return 'a'
        # todo is this slowdown?
        if not can_trace_type(value):
            value = str(type(value))
        if type(value) == dict:
            return value.copy()
        return value


class MethodEnterEvent(Event):
    cursor: ExecutionCursor
    input_variables: Variables
    stack: StackFrame

    def __init__(self, cursor: ExecutionCursor, stack: StackFrame, ts: float):
        self.cursor = cursor
        self.input_variables = Variables()
        self.stack = stack
        self.event_name = 'method_enter'
        self.ts = ts


class LineExecutionEvent(Event):
    cursor: ExecutionCursor
    locals: Variables
    stack: StackFrame

    def __init__(self, cursor, stack: StackFrame, ts: float):
        self.cursor = cursor
        self.locals = Variables()
        self.event_name = 'line'
        self.stack = stack
        self.ts = ts





class MethodExitEvent(Event):
    cursor: ExecutionCursor
    return_variables: Variables
    locals: Variables
    stack: StackFrame

    def __init__(self, cursor: ExecutionCursor, stack: StackFrame, ts: float):
        self.cursor = cursor
        self.return_variables = Variables()
        self.locals = Variables()
        self.event_name = 'method_exit'
        self.stack = stack
        self.ts = ts




from collections import deque, OrderedDict
from typing import Dict, Any

from pycrunch_trace.oop import Clock


class InlineProfiler:
    timings: Dict[str, float]

    def __init__(self):
        # method scope-> total time
        self.timings = OrderedDict()
        self.execution_stack = deque()

    def enter_scope(self, scope):
        self.execution_stack.append(scope)

    def get_full_stack(self):
        return '.'.join(self.execution_stack)

    def exit_scope(self):
        self.execution_stack.pop()

    def append_timing(self, scope: str, time_spent: float):
        stack___key = self.get_full_stack() + '.' + scope
        if not self.timings.get(stack___key):
            self.timings[stack___key] = time_spent
        else:
            self.timings[stack___key] += time_spent

    def print_timings(self):
        print(f'----print_timings----')
        for (key, val) in self.timings.items():
            print(f'{key}\t{val}')


inline_profiler_instance: InlineProfiler = InlineProfiler()
clock = Clock()

class ProfilingScope():
    def __init__(self, scope):
        self.scope = scope
        self.ts_begin = None
        self.ts_end = None

    def __enter__(self):
        self.ts_begin = clock.now()
        inline_profiler_instance.enter_scope(self.scope)
        return None

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.ts_end = clock.now()
        inline_profiler_instance.exit_scope()

        inline_profiler_instance.append_timing(self.scope, self.ts_end - self.ts_begin)


from pycrunch_trace.client.api.trace import Trace




def my_factorial(num: int):
    if num == 1:
        return num
    else:
        return num * my_factorial(num - 1)

yoba = Trace()
yoba.start(session_name='factorial')

my_factorial(10)

yoba.stop()
# print(x._tracer.simulation.simulated_code())


def find(symbol_to_find, in_str: str):
    accum = 0
    for s in in_str:
        if s == symbol_to_find:
            return accum
        accum += 1
    return -1

def kwar_testing(**kwargs):
    x = kwargs.get('x')
    y = kwargs.get('b')
    if not y:
        x = 2 + x
    something = 42


import io
import struct
from pathlib import Path

from . import tags


class TLV:
    length: int
    offset: int
    tag: int

    def __init__(self):
        self.length = None
        self.offset = None
        self.tag = None

    def data_offset(self):
        int_size = 4
        # TAG 4b | LEN 4b | <-----
        return self.offset + int_size * 2

class TraceFile:
    header: TLV
    file_section: TLV
    chunked_recording_filename = 'session.chunked.pycrunch-trace'

    def __init__(self, session_id: str, target_file: Path):
        self.target_file = target_file
        self.session_id = session_id
        # Header size may change in future
        self.header_size = 16 * 1024
        self.version_major = 0
        self.version_minor = 1

        self.header = None
        self.file_section = None


    def update_file_header_metadata_section(self, metadata_bytes_len):
        int_size = struct.calcsize(">i")
        uint64_size = struct.calcsize(">Q")

        bytes_written = self.target_file.stat().st_size

        with io.FileIO(self.target_file, 'r+') as file_to_write:
            header_metadata_begins_at = int_size * 2 + (uint64_size * 2)
            file_to_write.seek(self.header.data_offset())
            self.skip_to_free_header_chunk(file_to_write)

            # Q - unsigned = 8 bytes

            file_to_write.write(Int32(tags.HEADER_TAG_METADATA).bytes())
            file_to_write.write(Int32(uint64_size*2).bytes())

            file_to_write.write(Int64(bytes_written).bytes())
            file_to_write.write(Int64(metadata_bytes_len).bytes())

    def update_file_header_files_section(self, total_bytes):
        bytes_written = self.target_file.stat().st_size

        with io.FileIO(self.target_file, 'r+') as file_to_write:
            # Magic_Number | HEADER_SIZE | FILES_BEGINS | FILES_SIZE

            header_begins_at = self.header.data_offset()
            file_to_write.seek(header_begins_at)
            self.skip_to_free_header_chunk(file_to_write)

            # Q - unsigned = 8 bytes
            file_to_write.write(Int32(tags.HEADER_TAG_FILES).bytes())
            file_to_write.write(Int32(16).bytes())
            file_to_write.write(Int64(bytes_written).bytes())
            file_to_write.write(Int64(total_bytes).bytes())

    def skip_to_free_header_chunk(self, file_to_write):
        print('skip_to_free_header_chunk')
        print(f'pos = {file_to_write.tell()}')
        int_size = struct.calcsize(">i")
        has_data = True
        while has_data:
            buffer = file_to_write.read(int_size)
            tag = struct.unpack('>i', buffer)[0]
            if tag == 0:
                has_data = False
                # return cursor to free space
                file_to_write.seek(-int_size, io.SEEK_CUR)
                break
            buffer = file_to_write.read(int_size)
            next_payload_length = struct.unpack('>i', buffer)[0]
            # skip to next record
            file_to_write.seek(next_payload_length, io.SEEK_CUR)
        print(f'after pos = {file_to_write.tell()}')

    def write_header_placeholder(self):
        target_mode = 'w'
        with io.FileIO(self.target_file, target_mode) as file_to_write:
            self.write_signature(file_to_write)
            self.write_header(file_to_write)

    def write_signature(self, file_to_write):
        signature = Int32(15051991).bytes()
        file_to_write.write(signature)

    def write_header(self, file_to_write):
        self.header = TLV()
        self.header.offset = file_to_write.tell()

        # SIG | [ Tag | Len | Val ]

        self.header.length = self.header_size
        print(f'offse {self.header.offset}')
        print(f'l {self.header.length}')
        file_to_write.write(Int32(tags.TRACE_TAG_HEADER).bytes())
        file_to_write.write(Int32(self.header_size).bytes())

        file_to_write.write(Int32(tags.HEADER_TAG_VERSION).bytes())
        file_to_write.write(Int32(8).bytes())
        file_to_write.write(Int32(self.version_major).bytes())
        file_to_write.write(Int32(self.version_minor).bytes())

        ppp = file_to_write.seek(
            self.header.offset + 0 + 4 + 4 + self.header.length, io.SEEK_SET
        )
        print(f'hs={self.header_size}')
        print(f'ppp={ppp}')
        # this is to make sure file will have 16kb allocated at the beginning
        self.write_signature(file_to_write)

    def flush_chunk(self, tag_id: int, bytes_to_write):
        length_of_message = len(bytes_to_write)
        with io.FileIO(self.target_file, 'a') as file_to_write:
            # TLV triplet again
            file_to_write.write(Int32(tag_id).bytes())
            file_to_write.write(Int32(length_of_message).bytes())
            file_to_write.write(bytes_to_write)

class Int32:
    def __init__(self, value: int):
        self._value = value

    def bytes(self):
        return struct.pack(">i", self._value)

class Int64:
    def __init__(self, value: int):
        self._value = value

    def bytes(self):
        return struct.pack(">Q", self._value)



import setuptools
from distutils.core import setup

setup(name='pycrunch-trace',
      version='0.1.0',
      description='PyCrunch Tracing Client',
      url='http://github.com/gleb-sevruk/pycrunch-trace',
      author='Gleb Sevruk',
      author_email='gleb@pycrunch.com',
      license='libpng',
      keywords="tracing time travel debugging live coding",
      packages=setuptools.find_packages(),
      download_url='https://github.com/gleb-sevruk/pycrunch-trace/archive/v0.1.0.tar.gz',
      setup_requires=['wheel'],
      install_requires=[
          'jsonpickle',
          'PyYAML',
          'protobuf==3.11.3'
      ],
      include_package_data=True,
      zip_safe=False)



from typing import Dict, Any

import logging

from pycrunch_trace.server.trace_in_progress import TraceInProgress

logger = logging.getLogger(__name__)




class IncomingTraces:
    sessions_by_id: Dict[str, TraceInProgress]

    def __init__(self):
        self.sessions_by_id = dict()

    def trace_will_start(self, session_id):
        self.delete_possible_old_session(session_id)
        self._create_trace_in_progress(session_id)

    def _create_trace_in_progress(self, session_id):
        self.sessions_by_id[session_id] = TraceInProgress(session_id)

    def did_receive_more_events(self, session_id, events_count):
        current = self.get_session_with_id(session_id)
        current.add_events(events_count)

    def get_session_with_id(self, session_id) -> TraceInProgress:
        return self.sessions_by_id[session_id]

    def delete_possible_old_session(self, session_id):
        if session_id in self.sessions_by_id:
            self.log('Deleting stale/old session')
            del self.sessions_by_id[session_id]

    def log(self, msg):
        logger.info(msg)

incoming_traces: IncomingTraces = IncomingTraces()



from pycrunch_trace.client.api import trace


def nested_two(x):
    return x * 2

# down arrow (step over) should not enter this function
def nested_one(x):
    return x + 1


@trace
def demo_stepping_nested_calls():
    x = 1
    y = nested_one(nested_two(x))
    print(y)


demo_stepping_nested_calls()



from time import sleep

import requests

from pycrunch_trace.client.api.trace import Trace


def some_method():
    accum_counter = 1
    while accum_counter < 2:
        accum_counter += 1
        # try:
        r = requests.get('http://dev.iws.briteapps.space/')
        # r = requests.get('https://google.com/')
        # print(r.status_code)
        # print(r.content)
        x = accum_counter
        # except Exception as e:
        #     print("!!!!! FAIL", e)


yoba = Trace()
yoba.start('request_exce2',profile_name='default.profile.yaml')

some_method()

yoba.stop()
sleep(10)
# print(yoba.command_buffer)
# x = 1


from pycrunch_trace.server.perf import PerformanceInsights


def test_average_package_size_one_chunk():
    sut = PerformanceInsights()

    sut.sample(session_id='session_id', number_of_events=10, buffer_size=2000)

    actual = sut.average_event_size(session_id='session_id')
    assert actual == 200

def test_average_package_size_three_chunks():
    sut = PerformanceInsights()

    sut.sample(session_id='session_id', number_of_events=1, buffer_size=200)
    sut.sample(session_id='session_id', number_of_events=1, buffer_size=250)
    sut.sample(session_id='session_id', number_of_events=1, buffer_size=300)

    actual = sut.average_event_size(session_id='session_id')
    assert actual == 250

def test_average_package_size_three_sessions():
    sut = PerformanceInsights()

    sut.sample(session_id='1', number_of_events=1, buffer_size=200)
    sut.sample(session_id='2', number_of_events=1, buffer_size=250)
    sut.sample(session_id='3', number_of_events=2, buffer_size=700)

    assert 200 == sut.average_event_size(session_id='1')
    assert 250 == sut.average_event_size(session_id='2')
    assert 350 == sut.average_event_size(session_id='3')



from typing import List

from . import base_event
from .method_enter import MethodEnterEvent, MethodExitEvent, LineExecutionEvent
from ..file_system.human_readable_size import HumanReadableByteSize
from ..file_system.session_store import SessionStore

import pickle


def count_every_element(self, cleanup_function= None):
    accum = 0
    for e in self.buffer:
        if cleanup_function:
            cleanup_function(e)
        bytes_ = pickle.dumps(e)
        accum += len(bytes_)
    return accum

class SizeWithoutStack:
    def __init__(self, buffer):
        self.buffer = buffer

    def size(self):
        return count_every_element(self, self.clean_up_stack)

    def clean_up_stack(self, e):
        e.stack = None


class SizeOriginal:
    def __init__(self, buffer):
        self.buffer = buffer

    def size(self):
        return count_every_element(self)

    def clean_up_stack(self, e):
        e.stack = None


class SizeWithoutVariables:
    def __init__(self, buffer):
        self.buffer = buffer

    def size(self):
       return count_every_element(self, self.clean_up_vars)

    def clean_up_vars(self, e):
        if isinstance(e, MethodEnterEvent):
            e.input_variables = None
        if isinstance(e, MethodExitEvent):
            e.return_variables = None
            e.locals = None
        if isinstance(e, LineExecutionEvent):
            e.locals = None


class SizeWithoutCursor:
    def __init__(self, buffer):
        self.buffer = buffer

    def size(self):
        return count_every_element(self, self.clean_up_cursor)

    def clean_up_cursor(self, e):
        e.cursor = None

class SizeBreakdown:
    event_buffer: List[base_event.Event]

    @staticmethod
    def load_from_session():
        sess = SessionStore().load_session('request_exce')
        sess.load_metadata()
        print(f'metadata thinks size is: {sess.metadata.file_size_on_disk}')
        print()

        orig = SizeOriginal(sess.load_buffer())
        real_size = orig.size()
        SizeBreakdown.print_size('real size', real_size)

        total_bytes_so_far = SizeWithoutStack(sess.load_buffer())
        without_stack = total_bytes_so_far.size()
        SizeBreakdown.print_size('without stack', without_stack)

        without_variables = SizeWithoutVariables(sess.load_buffer())
        without_variables_size = without_variables.size()
        SizeBreakdown.print_size('without variables', without_variables_size)

        without_cursor = SizeWithoutCursor(sess.load_buffer())
        without_cursor_size = without_cursor.size()
        SizeBreakdown.print_size('without cursor', without_cursor_size)

        cursor = SizeWithoutCursor(sess.load_buffer())
        cursor.size()
        without_cursor_and_vars = SizeWithoutVariables(cursor.buffer)
        without_cursor_and_vars_size = without_cursor_and_vars.size()
        SizeBreakdown.print_size('without_cursor and vars', without_cursor_and_vars_size)

        print('matan:')


        # for i in range(100):
        #     print(bugger[i].event_name)

    @staticmethod
    def print_size(prefix, real_size):
        print(f'{prefix}: {HumanReadableByteSize(real_size)} ({real_size})')


import jsonpickle
from pycrunch.insights import trace

from pycrunch_trace.events import event_buffer_in_protobuf
from pycrunch_trace.filters import DefaultFileFilter
from pycrunch_trace.proto import message_pb2
from pycrunch_trace.serialization import to_string
from pycrunch_trace.tracing.simulation import models
from pycrunch_trace.tracing.simple_tracer import SimpleTracer
from pycrunch_trace.tracing.simulator_sink import SimulationEvent


def create_event_1():
    code_clone = models.Code()
    code_clone.co_name = 'my_factorial'
    code_clone.co_filename = '/Users/gleb/code/pycrunch_tracing/pycrunch_tracer/demo/factorial_demo.py'
    code_clone.co_argcount = 1

    sim_frame = models.Frame()
    sim_frame.f_lineno = 7
    sim_frame.f_locals = {
        'num': 10}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'call', None)

    return evt


def create_event_2():
    code_clone = models.Code()
    code_clone.co_name = 'my_factorial'
    code_clone.co_filename = '/Users/gleb/code/pycrunch_tracing/pycrunch_tracer/demo/factorial_demo.py'
    code_clone.co_argcount = 1

    sim_frame = models.Frame()
    sim_frame.f_lineno = 8
    sim_frame.f_locals = {
        'num': 10}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'line', None)

    return evt


def create_event_3():
    code_clone = models.Code()
    code_clone.co_name = 'my_factorial'
    code_clone.co_filename = '/Users/gleb/code/pycrunch_tracing/pycrunch_tracer/demo/factorial_demo.py'
    code_clone.co_argcount = 1

    sim_frame = models.Frame()
    sim_frame.f_lineno = 11
    sim_frame.f_locals = {
        'num': 10}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'line', None)

    return evt


def create_event_4():
    code_clone = models.Code()
    code_clone.co_name = 'my_factorial'
    code_clone.co_filename = '/Users/gleb/code/pycrunch_tracing/pycrunch_tracer/demo/factorial_demo.py'
    code_clone.co_argcount = 1

    sim_frame = models.Frame()
    sim_frame.f_lineno = 7
    sim_frame.f_locals = {
        'num': 9}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'call', None)

    return evt


def create_event_5():
    code_clone = models.Code()
    code_clone.co_name = 'my_factorial'
    code_clone.co_filename = '/Users/gleb/code/pycrunch_tracing/pycrunch_tracer/demo/factorial_demo.py'
    code_clone.co_argcount = 1

    sim_frame = models.Frame()
    sim_frame.f_lineno = 8
    sim_frame.f_locals = {
        'num': 9}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'line', None)

    return evt


def create_event_6():
    code_clone = models.Code()
    code_clone.co_name = 'my_factorial'
    code_clone.co_filename = '/Users/gleb/code/pycrunch_tracing/pycrunch_tracer/demo/factorial_demo.py'
    code_clone.co_argcount = 1

    sim_frame = models.Frame()
    sim_frame.f_lineno = 11
    sim_frame.f_locals = {
        'num': 9}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'line', None)

    return evt


def create_event_7():
    code_clone = models.Code()
    code_clone.co_name = 'my_factorial'
    code_clone.co_filename = '/Users/gleb/code/pycrunch_tracing/pycrunch_tracer/demo/factorial_demo.py'
    code_clone.co_argcount = 1

    sim_frame = models.Frame()
    sim_frame.f_lineno = 7
    sim_frame.f_locals = {
        'num': 8}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'call', None)

    return evt


def create_event_8():
    code_clone = models.Code()
    code_clone.co_name = 'my_factorial'
    code_clone.co_filename = '/Users/gleb/code/pycrunch_tracing/pycrunch_tracer/demo/factorial_demo.py'
    code_clone.co_argcount = 1

    sim_frame = models.Frame()
    sim_frame.f_lineno = 8
    sim_frame.f_locals = {
        'num': 8}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'line', None)

    return evt


def create_event_9():
    code_clone = models.Code()
    code_clone.co_name = 'my_factorial'
    code_clone.co_filename = '/Users/gleb/code/pycrunch_tracing/pycrunch_tracer/demo/factorial_demo.py'
    code_clone.co_argcount = 1

    sim_frame = models.Frame()
    sim_frame.f_lineno = 11
    sim_frame.f_locals = {
        'num': 8}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'line', None)

    return evt


def create_event_10():
    code_clone = models.Code()
    code_clone.co_name = 'my_factorial'
    code_clone.co_filename = '/Users/gleb/code/pycrunch_tracing/pycrunch_tracer/demo/factorial_demo.py'
    code_clone.co_argcount = 1

    sim_frame = models.Frame()
    sim_frame.f_lineno = 7
    sim_frame.f_locals = {
        'num': 7}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'call', None)

    return evt


def create_event_11():
    code_clone = models.Code()
    code_clone.co_name = 'my_factorial'
    code_clone.co_filename = '/Users/gleb/code/pycrunch_tracing/pycrunch_tracer/demo/factorial_demo.py'
    code_clone.co_argcount = 1

    sim_frame = models.Frame()
    sim_frame.f_lineno = 8
    sim_frame.f_locals = {
        'num': 7}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'line', None)

    return evt


def create_event_12():
    code_clone = models.Code()
    code_clone.co_name = 'my_factorial'
    code_clone.co_filename = '/Users/gleb/code/pycrunch_tracing/pycrunch_tracer/demo/factorial_demo.py'
    code_clone.co_argcount = 1

    sim_frame = models.Frame()
    sim_frame.f_lineno = 11
    sim_frame.f_locals = {
        'num': 7}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'line', None)

    return evt


def create_event_13():
    code_clone = models.Code()
    code_clone.co_name = 'my_factorial'
    code_clone.co_filename = '/Users/gleb/code/pycrunch_tracing/pycrunch_tracer/demo/factorial_demo.py'
    code_clone.co_argcount = 1

    sim_frame = models.Frame()
    sim_frame.f_lineno = 7
    sim_frame.f_locals = {
        'num': 6}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'call', None)

    return evt


def create_event_14():
    code_clone = models.Code()
    code_clone.co_name = 'my_factorial'
    code_clone.co_filename = '/Users/gleb/code/pycrunch_tracing/pycrunch_tracer/demo/factorial_demo.py'
    code_clone.co_argcount = 1

    sim_frame = models.Frame()
    sim_frame.f_lineno = 8
    sim_frame.f_locals = {
        'num': 6}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'line', None)

    return evt


def create_event_15():
    code_clone = models.Code()
    code_clone.co_name = 'my_factorial'
    code_clone.co_filename = '/Users/gleb/code/pycrunch_tracing/pycrunch_tracer/demo/factorial_demo.py'
    code_clone.co_argcount = 1

    sim_frame = models.Frame()
    sim_frame.f_lineno = 11
    sim_frame.f_locals = {
        'num': 6}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'line', None)

    return evt


def create_event_16():
    code_clone = models.Code()
    code_clone.co_name = 'my_factorial'
    code_clone.co_filename = '/Users/gleb/code/pycrunch_tracing/pycrunch_tracer/demo/factorial_demo.py'
    code_clone.co_argcount = 1

    sim_frame = models.Frame()
    sim_frame.f_lineno = 7
    sim_frame.f_locals = {
        'num': 5}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'call', None)

    return evt


def create_event_17():
    code_clone = models.Code()
    code_clone.co_name = 'my_factorial'
    code_clone.co_filename = '/Users/gleb/code/pycrunch_tracing/pycrunch_tracer/demo/factorial_demo.py'
    code_clone.co_argcount = 1

    sim_frame = models.Frame()
    sim_frame.f_lineno = 8
    sim_frame.f_locals = {
        'num': 5}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'line', None)

    return evt


def create_event_18():
    code_clone = models.Code()
    code_clone.co_name = 'my_factorial'
    code_clone.co_filename = '/Users/gleb/code/pycrunch_tracing/pycrunch_tracer/demo/factorial_demo.py'
    code_clone.co_argcount = 1

    sim_frame = models.Frame()
    sim_frame.f_lineno = 11
    sim_frame.f_locals = {
        'num': 5}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'line', None)

    return evt


def create_event_19():
    code_clone = models.Code()
    code_clone.co_name = 'my_factorial'
    code_clone.co_filename = '/Users/gleb/code/pycrunch_tracing/pycrunch_tracer/demo/factorial_demo.py'
    code_clone.co_argcount = 1

    sim_frame = models.Frame()
    sim_frame.f_lineno = 7
    sim_frame.f_locals = {
        'num': 4}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'call', None)

    return evt


def create_event_20():
    code_clone = models.Code()
    code_clone.co_name = 'my_factorial'
    code_clone.co_filename = '/Users/gleb/code/pycrunch_tracing/pycrunch_tracer/demo/factorial_demo.py'
    code_clone.co_argcount = 1

    sim_frame = models.Frame()
    sim_frame.f_lineno = 8
    sim_frame.f_locals = {
        'num': 4}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'line', None)

    return evt


def create_event_21():
    code_clone = models.Code()
    code_clone.co_name = 'my_factorial'
    code_clone.co_filename = '/Users/gleb/code/pycrunch_tracing/pycrunch_tracer/demo/factorial_demo.py'
    code_clone.co_argcount = 1

    sim_frame = models.Frame()
    sim_frame.f_lineno = 11
    sim_frame.f_locals = {
        'num': 4}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'line', None)

    return evt


def create_event_22():
    code_clone = models.Code()
    code_clone.co_name = 'my_factorial'
    code_clone.co_filename = '/Users/gleb/code/pycrunch_tracing/pycrunch_tracer/demo/factorial_demo.py'
    code_clone.co_argcount = 1

    sim_frame = models.Frame()
    sim_frame.f_lineno = 7
    sim_frame.f_locals = {
        'num': 3}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'call', None)

    return evt


def create_event_23():
    code_clone = models.Code()
    code_clone.co_name = 'my_factorial'
    code_clone.co_filename = '/Users/gleb/code/pycrunch_tracing/pycrunch_tracer/demo/factorial_demo.py'
    code_clone.co_argcount = 1

    sim_frame = models.Frame()
    sim_frame.f_lineno = 8
    sim_frame.f_locals = {
        'num': 3}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'line', None)

    return evt


def create_event_24():
    code_clone = models.Code()
    code_clone.co_name = 'my_factorial'
    code_clone.co_filename = '/Users/gleb/code/pycrunch_tracing/pycrunch_tracer/demo/factorial_demo.py'
    code_clone.co_argcount = 1

    sim_frame = models.Frame()
    sim_frame.f_lineno = 11
    sim_frame.f_locals = {
        'num': 3}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'line', None)

    return evt


def create_event_25():
    code_clone = models.Code()
    code_clone.co_name = 'my_factorial'
    code_clone.co_filename = '/Users/gleb/code/pycrunch_tracing/pycrunch_tracer/demo/factorial_demo.py'
    code_clone.co_argcount = 1

    sim_frame = models.Frame()
    sim_frame.f_lineno = 7
    sim_frame.f_locals = {
        'num': 2}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'call', None)

    return evt


def create_event_26():
    code_clone = models.Code()
    code_clone.co_name = 'my_factorial'
    code_clone.co_filename = '/Users/gleb/code/pycrunch_tracing/pycrunch_tracer/demo/factorial_demo.py'
    code_clone.co_argcount = 1

    sim_frame = models.Frame()
    sim_frame.f_lineno = 8
    sim_frame.f_locals = {
        'num': 2}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'line', None)

    return evt


def create_event_27():
    code_clone = models.Code()
    code_clone.co_name = 'my_factorial'
    code_clone.co_filename = '/Users/gleb/code/pycrunch_tracing/pycrunch_tracer/demo/factorial_demo.py'
    code_clone.co_argcount = 1

    sim_frame = models.Frame()
    sim_frame.f_lineno = 11
    sim_frame.f_locals = {
        'num': 2}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'line', None)

    return evt


def create_event_28():
    code_clone = models.Code()
    code_clone.co_name = 'my_factorial'
    code_clone.co_filename = '/Users/gleb/code/pycrunch_tracing/pycrunch_tracer/demo/factorial_demo.py'
    code_clone.co_argcount = 1

    sim_frame = models.Frame()
    sim_frame.f_lineno = 7
    sim_frame.f_locals = {
        'num': 1}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'call', None)

    return evt


def create_event_29():
    code_clone = models.Code()
    code_clone.co_name = 'my_factorial'
    code_clone.co_filename = '/Users/gleb/code/pycrunch_tracing/pycrunch_tracer/demo/factorial_demo.py'
    code_clone.co_argcount = 1

    sim_frame = models.Frame()
    sim_frame.f_lineno = 8
    sim_frame.f_locals = {
        'num': 1}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'line', None)

    return evt


def create_event_30():
    code_clone = models.Code()
    code_clone.co_name = 'my_factorial'
    code_clone.co_filename = '/Users/gleb/code/pycrunch_tracing/pycrunch_tracer/demo/factorial_demo.py'
    code_clone.co_argcount = 1

    sim_frame = models.Frame()
    sim_frame.f_lineno = 9
    sim_frame.f_locals = {
        'num': 1}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'line', None)

    return evt


def create_event_31():
    code_clone = models.Code()
    code_clone.co_name = 'my_factorial'
    code_clone.co_filename = '/Users/gleb/code/pycrunch_tracing/pycrunch_tracer/demo/factorial_demo.py'
    code_clone.co_argcount = 1

    sim_frame = models.Frame()
    sim_frame.f_lineno = 9
    sim_frame.f_locals = {
        'num': 1}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'return', 1)

    return evt


def create_event_32():
    code_clone = models.Code()
    code_clone.co_name = 'my_factorial'
    code_clone.co_filename = '/Users/gleb/code/pycrunch_tracing/pycrunch_tracer/demo/factorial_demo.py'
    code_clone.co_argcount = 1

    sim_frame = models.Frame()
    sim_frame.f_lineno = 11
    sim_frame.f_locals = {
        'num': 2}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'return', 2)

    return evt


def create_event_33():
    code_clone = models.Code()
    code_clone.co_name = 'my_factorial'
    code_clone.co_filename = '/Users/gleb/code/pycrunch_tracing/pycrunch_tracer/demo/factorial_demo.py'
    code_clone.co_argcount = 1

    sim_frame = models.Frame()
    sim_frame.f_lineno = 11
    sim_frame.f_locals = {
        'num': 3}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'return', 6)

    return evt


def create_event_34():
    code_clone = models.Code()
    code_clone.co_name = 'my_factorial'
    code_clone.co_filename = '/Users/gleb/code/pycrunch_tracing/pycrunch_tracer/demo/factorial_demo.py'
    code_clone.co_argcount = 1

    sim_frame = models.Frame()
    sim_frame.f_lineno = 11
    sim_frame.f_locals = {
        'num': 4}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'return', 24)

    return evt


def create_event_35():
    code_clone = models.Code()
    code_clone.co_name = 'my_factorial'
    code_clone.co_filename = '/Users/gleb/code/pycrunch_tracing/pycrunch_tracer/demo/factorial_demo.py'
    code_clone.co_argcount = 1

    sim_frame = models.Frame()
    sim_frame.f_lineno = 11
    sim_frame.f_locals = {
        'num': 5}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'return', 120)

    return evt


def create_event_36():
    code_clone = models.Code()
    code_clone.co_name = 'my_factorial'
    code_clone.co_filename = '/Users/gleb/code/pycrunch_tracing/pycrunch_tracer/demo/factorial_demo.py'
    code_clone.co_argcount = 1

    sim_frame = models.Frame()
    sim_frame.f_lineno = 11
    sim_frame.f_locals = {
        'num': 6}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'return', 720)

    return evt


def create_event_37():
    code_clone = models.Code()
    code_clone.co_name = 'my_factorial'
    code_clone.co_filename = '/Users/gleb/code/pycrunch_tracing/pycrunch_tracer/demo/factorial_demo.py'
    code_clone.co_argcount = 1

    sim_frame = models.Frame()
    sim_frame.f_lineno = 11
    sim_frame.f_locals = {
        'num': 7}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'return', 5040)

    return evt


def create_event_38():
    code_clone = models.Code()
    code_clone.co_name = 'my_factorial'
    code_clone.co_filename = '/Users/gleb/code/pycrunch_tracing/pycrunch_tracer/demo/factorial_demo.py'
    code_clone.co_argcount = 1

    sim_frame = models.Frame()
    sim_frame.f_lineno = 11
    sim_frame.f_locals = {
        'num': 8}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'return', 40320)

    return evt


def create_event_39():
    code_clone = models.Code()
    code_clone.co_name = 'my_factorial'
    code_clone.co_filename = '/Users/gleb/code/pycrunch_tracing/pycrunch_tracer/demo/factorial_demo.py'
    code_clone.co_argcount = 1

    sim_frame = models.Frame()
    sim_frame.f_lineno = 11
    sim_frame.f_locals = {
        'num': 9}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'return', 362880)

    return evt


def create_event_40():
    code_clone = models.Code()
    code_clone.co_name = 'my_factorial'
    code_clone.co_filename = '/Users/gleb/code/pycrunch_tracing/pycrunch_tracer/demo/factorial_demo.py'
    code_clone.co_argcount = 1

    sim_frame = models.Frame()
    sim_frame.f_lineno = 11
    sim_frame.f_locals = {
        'num': 10}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'return', 3628800)

    return evt


def test_simulated():
    events = []
    events.append(create_event_1())
    events.append(create_event_2())
    events.append(create_event_3())
    events.append(create_event_4())
    events.append(create_event_5())
    events.append(create_event_6())
    events.append(create_event_7())
    events.append(create_event_8())
    events.append(create_event_9())
    events.append(create_event_10())
    events.append(create_event_11())
    events.append(create_event_12())
    events.append(create_event_13())
    events.append(create_event_14())
    events.append(create_event_15())
    events.append(create_event_16())
    events.append(create_event_17())
    events.append(create_event_18())
    events.append(create_event_19())
    events.append(create_event_20())
    events.append(create_event_21())
    events.append(create_event_22())
    events.append(create_event_23())
    events.append(create_event_24())
    events.append(create_event_25())
    events.append(create_event_26())
    events.append(create_event_27())
    events.append(create_event_28())
    events.append(create_event_29())
    events.append(create_event_30())
    events.append(create_event_31())
    events.append(create_event_32())
    events.append(create_event_33())
    events.append(create_event_34())
    events.append(create_event_35())
    events.append(create_event_36())
    events.append(create_event_37())
    events.append(create_event_38())
    events.append(create_event_39())
    events.append(create_event_40())

    event_buffer = []
    sut = SimpleTracer(event_buffer, 'sim_round', DefaultFileFilter(),)
    for x in events:
        sut.simple_tracer(x.frame, x.event, x.arg)

    x2 = to_string(event_buffer)
    trace(x2=x2)

    array_round = jsonpickle.decode(x2)
    # trace(array_round=array_round)
    # ss = SessionStore()
    # ses = ss.new_session('round_demo')
    # ses.save_with_metadata()

    bytesAsString = event_buffer_in_protobuf.EventBufferInProtobuf(event_buffer, x.file_map).as_bytes()
    trace(vbbbb=str(bytesAsString))
    print(len(bytesAsString))
    session_round = message_pb2.TraceSession()

    x = session_round.ParseFromString(bytesAsString)
    print(x)